/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:                                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996,97)                         */
/*                                                                          */
/****************************************************************************/
#include "albert.h"

#if DIM != DIM_OF_WORLD
#error not now for DIM != DIM_OF_WORLD
#endif

/*--------------------------------------------------------------------------*/
/*  compute volume of an element                                            */
/*--------------------------------------------------------------------------*/

REAL el_det(const EL_INFO *el_info)
{
  FUNCNAME("el_det");

  TEST_FLAG(FILL_COORDS, el_info);
  return(el_info->coord[1][0] - el_info->coord[0][0]);
}

REAL el_volume(const EL_INFO *el_info)
{
  return(el_det(el_info));
}

/*--------------------------------------------------------------------------*/
/*  compute gradients of basis functions on element; return the absulute    */
/*  value of the determinante from the transformation to the reference      */
/*  element                                                                 */
/*--------------------------------------------------------------------------*/

REAL el_grd_lambda(const EL_INFO *el_info, REAL grd_lam[DIM+1][DIM_OF_WORLD])
{
  FUNCNAME("el_grd_lambda");
  REAL   det;

  TEST_FLAG(FILL_COORDS, el_info);

  det = el_info->coord[1][0] - el_info->coord[0][0];
  grd_lam[0][0] = -1.0/det;
  grd_lam[1][0] =  1.0/det;

  return(det);
}

#if DIM_OF_WORLD == 3
/****************************************************************************/
/*  vector product in R^3                                                   */
/****************************************************************************/

void vector_product(const REAL_D x, const REAL_D y, REAL_D z)
{
  FUNCNAME("vector_product");

  z[0] = x[1]*y[2] - x[2]*y[1];
  z[1] = x[2]*y[0] - x[0]*y[2];
  z[2] = x[0]*y[1] - x[1]*y[0];

  return;
}
#endif
