#include "albert.h"
#include <glrnd.h>
#include <glmesh.h>
static void xmin_xmax(MESH *mesh, REAL_D xmin, REAL_D xmax);

#if DIM == 1
#define GL_ELEMENTS_0  1
#define GL_COORDS_0    N_VERTICES
#define GL_ELEMENTS_1  1
#define GL_COORDS_1    N_VERTICES
#define GL_ELEMENTS_2  2
#define GL_COORDS_2    N_VERTICES + 1
#define GL_ELEMENTS_3  3
#define GL_COORDS_3    N_VERTICES + 2
#define GL_ELEMENTS_4  4
#define GL_COORDS_4    N_VERTICES + 3
#define GL_ELEMENTS_42 2
#define GL_COORDS_42   N_VERTICES + 1

#define GL_MAX_COORD   GL_COORDS_4

/****************************************************************************/
/*  local number of DOFs and vertices on the sub elements                   */
/****************************************************************************/

static int nc0[GL_ELEMENTS_0][2] = {{0,1}};
static int nc1[GL_ELEMENTS_1][2] = {{0,1}};
static int nc2[GL_ELEMENTS_2][2] = {{0,2},{2,1}};
static int nc3[GL_ELEMENTS_3][2] = {{0,2},{2,3},{3,1}};
static int nc4[GL_ELEMENTS_4][2] = {{0,2},{2,3},{3,4},{4,1}};
static int nc42[GL_ELEMENTS_42][2] = {{0,3},{3,1}};

/****************************************************************************/
/*  functions for calculating the coordinates at the Lagrange nodes         */
/****************************************************************************/

static void lagrange_2(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
    p[2][i] = 0.5*(coord[0][i] + coord[1][i]);
  return;
}

static void lagrange_3(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
  {
    p[2][i] = 2.0/3.0*coord[0][i] + 1.0/3.0*coord[1][i];
    p[3][i] = 1.0/3.0*coord[0][i] + 2.0/3.0*coord[1][i];
  }
  return;
}

static void lagrange_4(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
  {
    p[2][i] = 0.75*coord[0][i] + 0.25*coord[1][i];
    p[3][i] = 0.50*coord[0][i] + 0.50*coord[1][i];
    p[4][i] = 0.25*coord[0][i] + 0.75*coord[1][i];
  }
  return;
}

static void lagrange_42(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
    p[2][i]  = 0.5*(coord[0][i] + coord[1][i]);
  return;
}
#endif

#if DIM == 2
#define GL_ELEMENTS_0  1
#define GL_COORDS_0    N_VERTICES
#define GL_ELEMENTS_1  1
#define GL_COORDS_1    N_VERTICES
#define GL_ELEMENTS_2  4
#define GL_COORDS_2    N_VERTICES + N_EDGES
#define GL_ELEMENTS_3  9
#define GL_COORDS_3    N_VERTICES + 2*N_EDGES + 1
#define GL_ELEMENTS_4  16
#define GL_COORDS_4    N_VERTICES + 3*N_EDGES + 3
#define GL_ELEMENTS_42 4
#define GL_COORDS_42   N_VERTICES + N_EDGES

#define GL_MAX_COORD   GL_COORDS_4

/****************************************************************************/
/*  local number of DOFs and vertices on the sub elements                   */
/****************************************************************************/

static int nc0[GL_ELEMENTS_0][3] = {{0,1,2}};
static int nc1[GL_ELEMENTS_1][3] = {{0,1,2}};
static int nc2[GL_ELEMENTS_2][3] = {{0,5,4},{5,1,3},{4,3,2},{3,4,5}};
static int nc3[GL_ELEMENTS_3][3] = {{0,7,6},{8,1,3},{5,4,2},{7,8,9},
				    {9,3,4},{6,9,5},{9,6,7},{3,9,8},{4,5,9}};
static int nc4[GL_ELEMENTS_4][3] = {{ 0, 9, 8},{11, 1, 3},{ 6, 5, 2},
				    { 9,10,12},{10,11,13},{ 8,12, 7},
				    {12,13,14},{13, 3, 4},{ 7,14, 6},
				    {14, 4, 5},{12, 8, 9},{13,12,10},
				    { 3,13,11},{14, 7,12},{ 4,14,13},
				    { 5, 6,14}};
static int  nc42[GL_ELEMENTS_42][3] = {{0,10, 7},{10,1,4},{7,4,2},{4,7,10}};

/****************************************************************************/
/*  functions for calculating the coordinates at the Lagrange nodes         */
/****************************************************************************/

static void lagrange_2(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
  {
    p[3][i] = 0.5*(coord[1][i] + coord[2][i]);
    p[4][i] = 0.5*(coord[2][i] + coord[0][i]);
    p[5][i] = 0.5*(coord[0][i] + coord[1][i]);
  }
  return;
}

static void lagrange_3(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
  {
    p[3][i] = 2.0/3.0*coord[1][i] + 1.0/3.0*coord[2][i];
    p[4][i] = 1.0/3.0*coord[1][i] + 2.0/3.0*coord[2][i];
    p[5][i] = 1.0/3.0*coord[0][i] + 2.0/3.0*coord[2][i];
    p[6][i] = 2.0/3.0*coord[0][i] + 1.0/3.0*coord[2][i];
    p[7][i] = 2.0/3.0*coord[0][i] + 1.0/3.0*coord[1][i];
    p[8][i] = 1.0/3.0*coord[0][i] + 2.0/3.0*coord[1][i];
    p[9][i] = 1.0/3.0*(coord[0][i] + coord[1][i] + coord[2][i]);
  }
  return;
}

static void lagrange_4(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
  {
    p[3][i] = 0.75*coord[1][i] + 0.25*coord[2][i];
    p[4][i] = 0.5*(coord[1][i] + coord[2][i]);
    p[5][i] = 0.25*coord[1][i] + 0.75*coord[2][i];
    p[6][i] = 0.25*coord[0][i] + 0.75*coord[2][i];
    p[7][i] = 0.5*(coord[0][i] + coord[2][i]);
    p[8][i] = 0.75*coord[0][i] + 0.25*coord[2][i];
    p[9][i] = 0.75*coord[0][i] + 0.25*coord[1][i];
    p[10][i] = 0.5*(coord[0][i] + coord[1][i]);
    p[11][i] = 0.25*coord[0][i] + 0.75*coord[1][i];
    p[12][i] = 0.5*coord[0][i] + 0.25*(coord[1][i] + coord[2][i]);
    p[13][i] = 0.5*coord[1][i] + 0.25*(coord[0][i] + coord[2][i]);
    p[14][i] = 0.5*coord[2][i] + 0.25*(coord[0][i] + coord[1][i]);
  }
  return;
}

static void lagrange_42(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
  {
    p[4][i]  = 0.5*(coord[1][i] + coord[2][i]);
    p[7][i]  = 0.5*(coord[0][i] + coord[2][i]);
    p[10][i] = 0.5*(coord[0][i] + coord[1][i]);
  }
  return;
}
#endif

#if DIM == 3
#define GL_ELEMENTS_0  1
#define GL_COORDS_0    N_VERTICES
#define GL_ELEMENTS_1  1
#define GL_COORDS_1    N_VERTICES
#define GL_ELEMENTS_2  8
#define GL_COORDS_2    N_VERTICES + N_EDGES
#define GL_ELEMENTS_3  27
#define GL_COORDS_3    N_VERTICES + 2*N_EDGES + N_FACES
#define GL_ELEMENTS_4  64
#define GL_COORDS_4    N_VERTICES + 3*N_EDGES + 3*N_FACES + 1
#define GL_ELEMENTS_42 8
#define GL_COORDS_42   N_VERTICES + N_EDGES

#define GL_MAX_COORD   GL_COORDS_4

/****************************************************************************/
/*  local number of DOFs and vertices on the sub elements                   */
/****************************************************************************/

static int nc0[GL_ELEMENTS_0][4] = {{0,1,2,3}};
static int nc1[GL_ELEMENTS_1][4] = {{0,1,2,3}};
static int nc2[GL_ELEMENTS_2][4] = {{0,4,5,6},{4,1,7,8},{5,7,2,9},{6,8,9,3},
				    {6,4,8,5},{4,7,5,8},{7,9,8,5},{9,6,5,8}};

#if 0
static int nc3[GL_ELEMENTS_3][4] = {{0,7,6},{8,1,3},{5,4,2},{7,8,9},
				    {9,3,4},{6,9,5},{9,6,7},{3,9,8},{4,5,9}};
#endif
static int nc4[GL_ELEMENTS_4][4] = {{ 0, 4, 7,10},{ 4, 5,31,28},
				    { 7,31, 8,25},{10,28,25,11},
				    {10, 4,28, 7},{ 4,31, 7,28},
				    {31,25,28, 7},{25,10, 7,28},
				    { 5, 6,32,29},{ 6, 1,13,16},
				    {32,13,14,22},{29,16,22,17},
				    {29, 6,16,32},{ 6,13,32,16},
				    {13,22,16,32},{22,29,32,16},
				    { 8,33, 9,26},{33,14,15,23},
				    { 9,15, 2,19},{26,23,19,20},
				    {26,33,23, 9},{33,15, 9,23},
				    {15,19,23, 9},{19,26, 9,23},
				    {11,30,27,12},{30,17,24,18},
				    {27,24,20,21},{12,18,21, 3},
				    {12,30,18,27},{30,24,27,18},
				    {24,21,18,27},{21,12,27,18},
				    {11,28,30,25},{28, 5,29,31},
				    {30,29,17,34},{25,31,34, 8},
				    {25,28,31,30},{28,29,30,31},
				    {29,34,31,30},{34,25,30,31},
				    { 5,32,31,29},{32,14,33,22},
				    {31,33, 8,34},{29,22,34,17},
				    {29,32,22,31},{32,33,31,22},
				    {33,34,22,31},{34,29,31,22},
				    {14,23,22,33},{23,20,24,26},
				    {22,24,17,34},{33,26,34, 8},
				    {33,23,26,22},{23,24,22,26},
				    {24,34,26,22},{34,33,22,26},
				    {20,27,26,24},{27,11,25,30},
				    {26,25, 8,34},{24,30,34,17},
				    {24,27,30,26},{27,25,26,30},
				    {25,34,30,26},{34,24,26,30}};
static int  nc42[GL_ELEMENTS_42][4] = {{ 0, 5, 8,11},{ 5, 1,14,17},
				       { 8,14, 2,20},{11,17,20, 3},
				       {11, 5,17, 8},{ 5,14, 8,17},
				       {14,20,17, 8},{20,11, 8,17}};

/****************************************************************************/
/*  functions for calculating the coordinates at the Lagrange nodes         */
/****************************************************************************/

static void lagrange_2(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
  {
    p[4][i] = 0.5*(coord[0][i] + coord[1][i]);
    p[5][i] = 0.5*(coord[0][i] + coord[2][i]);
    p[6][i] = 0.5*(coord[0][i] + coord[3][i]);
    p[7][i] = 0.5*(coord[1][i] + coord[2][i]);
    p[8][i] = 0.5*(coord[1][i] + coord[3][i]);
    p[9][i] = 0.5*(coord[2][i] + coord[3][i]);
  }
  return;
}


static void lagrange_4(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
  {
    p[4][i]  = 0.75*coord[0][i] + 0.25*coord[1][i];
    p[5][i]  = 0.5*(coord[0][i] + coord[1][i]);
    p[6][i]  = 0.25*coord[0][i] + 0.75*coord[1][i];

    p[7][i]  = 0.75*coord[0][i] + 0.25*coord[2][i];
    p[8][i]  = 0.5*(coord[0][i] + coord[2][i]);
    p[9][i]  = 0.25*coord[0][i] + 0.75*coord[2][i];

    p[10][i] = 0.75*coord[0][i] + 0.25*coord[3][i];
    p[11][i] = 0.5*(coord[0][i] + coord[3][i]);
    p[12][i] = 0.25*coord[0][i] + 0.75*coord[3][i];

    p[13][i] = 0.75*coord[1][i] + 0.25*coord[2][i];
    p[14][i] = 0.5*(coord[1][i] + coord[2][i]);
    p[15][i] = 0.25*coord[1][i] + 0.75*coord[2][i];

    p[16][i] = 0.75*coord[1][i] + 0.25*coord[3][i];
    p[17][i] = 0.5*(coord[1][i] + coord[3][i]);
    p[18][i] = 0.25*coord[1][i] + 0.75*coord[3][i];

    p[19][i] = 0.75*coord[2][i] + 0.25*coord[3][i];
    p[20][i] = 0.5*(coord[2][i] + coord[3][i]);
    p[21][i] = 0.25*coord[2][i] + 0.75*coord[3][i];

    p[22][i] = 0.5*coord[1][i] + 0.25*(coord[2][i] + coord[3][i]);
    p[23][i] = 0.5*coord[2][i] + 0.25*(coord[1][i] + coord[3][i]);
    p[24][i] = 0.5*coord[3][i] + 0.25*(coord[1][i] + coord[2][i]);

    p[25][i] = 0.5*coord[0][i] + 0.25*(coord[2][i] + coord[3][i]);
    p[26][i] = 0.5*coord[2][i] + 0.25*(coord[0][i] + coord[3][i]);
    p[27][i] = 0.5*coord[3][i] + 0.25*(coord[0][i] + coord[2][i]);

    p[28][i] = 0.5*coord[0][i] + 0.25*(coord[1][i] + coord[3][i]);
    p[29][i] = 0.5*coord[1][i] + 0.25*(coord[0][i] + coord[3][i]);
    p[30][i] = 0.5*coord[3][i] + 0.25*(coord[0][i] + coord[1][i]);

    p[31][i] = 0.5*coord[0][i] + 0.25*(coord[1][i] + coord[2][i]);
    p[32][i] = 0.5*coord[1][i] + 0.25*(coord[0][i] + coord[2][i]);
    p[33][i] = 0.5*coord[2][i] + 0.25*(coord[0][i] + coord[1][i]);

    p[34][i] = 0.25*(coord[0][i] + coord[1][i] + coord[2][i] + coord[3][i]);
  }
  return;
}

static void lagrange_42(const REAL_D *coord, REAL **p)
{
  int  i;

  for(i = 0; i < DIM_OF_WORLD; i++)
  {
    p[5][i]  = 0.5*(coord[0][i] + coord[1][i]);
    p[8][i]  = 0.5*(coord[0][i] + coord[2][i]);
    p[11][i] = 0.5*(coord[0][i] + coord[3][i]);
    p[14][i] = 0.5*(coord[1][i] + coord[2][i]);
    p[17][i] = 0.5*(coord[1][i] + coord[3][i]);
    p[20][i] = 0.5*(coord[2][i] + coord[3][i]);
  }
  return;
}
#endif

/****************************************************************************/
/*  vector for storing the additional coordinates, and vector with pointers */
/*  to all these coordinates; the first three have to be set on each        */
/*  element to the element's verices                                        */
/****************************************************************************/

static REAL_D p[GL_MAX_COORD-N_VERTICES];
static REAL   *gl_coords[GL_MAX_COORD];

typedef struct gltools_info  GLTOOLS_INFO;
struct gltools_info
{
  void           *data;
  DOF_REAL_VEC   *u;
  int            degree;
  int            refinement;

  int            gl_elements;
  int            gl_coords;
  
  int            (*nc)[N_VERTICES];
  void           (*lagrange)(const REAL_D *coord, REAL **p);

  double         min, max;

  GLTOOLS_INFO   *next;
};

static GLTOOLS_INFO *key_P_info = nil;

static int glrKeyAction_P(glRenderer rnd, int mask)
{
  switch (key_P_info->degree)
  {
  case 0:
  case 1:
    return(GL_FALSE);
  case 2:
    if (key_P_info->refinement)
    {
      key_P_info->gl_elements = 1;
      key_P_info->gl_coords   = N_VERTICES;
      key_P_info->nc          = nc1;
      key_P_info->lagrange    = nil;
      key_P_info->refinement  = 0;
    }
    else
    {
      key_P_info->gl_elements = GL_ELEMENTS_2;
      key_P_info->gl_coords = GL_COORDS_2;
      key_P_info->nc = nc2;
      key_P_info->lagrange = lagrange_2;
      key_P_info->refinement  = 1;
    }
    return(GL_TRUE);
  case 3:
#if DIM <= 2
    if (key_P_info->refinement)
    {
      key_P_info->gl_elements = 1;
      key_P_info->gl_coords   = N_VERTICES;
      key_P_info->nc          = nc1;
      key_P_info->lagrange    = nil;
      key_P_info->refinement  = 0;
    }
    else
    {
      key_P_info->gl_elements = GL_ELEMENTS_3;
      key_P_info->gl_coords = GL_COORDS_3;
      key_P_info->nc = nc3;
      key_P_info->lagrange = lagrange_3;
      key_P_info->refinement  = 1;
    }
    return(GL_TRUE);
#elif DIM == 3
    return(GL_FALSE);
#endif
  case 4:
    if (key_P_info->refinement == 2)
    {
      key_P_info->gl_elements = 1;
      key_P_info->gl_coords   = N_VERTICES;
      key_P_info->nc          = nc1;
      key_P_info->lagrange    = nil;
      key_P_info->refinement  = 0;
    }
    else if (key_P_info->refinement == 1)
    {
      key_P_info->gl_elements = GL_ELEMENTS_4;
      key_P_info->gl_coords = GL_COORDS_4;
      key_P_info->nc = nc4;
      key_P_info->lagrange = lagrange_4;
      key_P_info->refinement  = 2;
    }
    else
    {
      key_P_info->gl_elements = GL_ELEMENTS_42;
      key_P_info->gl_coords = GL_COORDS_42;
      key_P_info->nc = nc42;
      key_P_info->lagrange = lagrange_42;
      key_P_info->refinement  = 1;
    }
    return(GL_TRUE);

    return(GL_TRUE);
  default:
    return(GL_FALSE);
  }
}

static void m_cell_loop(glMesh m, void *data, glmSimplexCallback sxf)
{
  FUNCNAME("m_cell_loop");
  const DOF        *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  void             (*lagrange)(const REAL_D *coord, REAL **p);
  EL_INFO          *el_info;
  REAL             *c[N_VERTICES];
  int              sub_dof[N_VERTICES], *dof;
  int              i, ic, n;
  GLTOOLS_INFO     *info = (struct gltools_info *) data;
  DOF_REAL_VEC     *u    = info->data;
  MESH             *mesh = u->fe_space->mesh;
  const DOF_ADMIN  *admin = u->fe_space->admin;
  TRAVERSE_STACK   *stack = get_traverse_stack();
  int              (*nc)[N_VERTICES] = info->nc;
  int              gl_elements = info->gl_elements;

  INFO(0,2)("\n");
  get_dof_indices = u->fe_space->bas_fcts->get_dof_indices;
  lagrange        = info->lagrange;

  ic=1;
  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS);
  while(el_info)
  {
    dof = (int *) get_dof_indices(el_info->el, admin, nil);

    for(i = 0; i < N_VERTICES;i++)
      gl_coords[i] = el_info->coord[i];

    if (info->degree > 0)
    {
      if (lagrange)  lagrange((void *) el_info->coord, gl_coords);

/****************************************************************************/
/*  now, display on the sub triangles                                       */
/****************************************************************************/

      for (n = 0; n < gl_elements; n++)
      {
	for (i = 0; i < N_VERTICES; i++)
	{
	  c[i]       = gl_coords[nc[n][i]];
	  sub_dof[i] = dof[nc[n][i]];
	}
	sxf(m, ic, 1, NULL, sub_dof, c);
      }
    }
    else
    {
      for (i = 0; i < N_VERTICES; i++)
      {
        c[i]       = gl_coords[i];
        sub_dof[i] = dof[0];
      }
      sxf(m, ic, 1, NULL, sub_dof, c);
    }
    ic++;
    el_info=traverse_next(stack,el_info);
  }

  free_traverse_stack(stack);

  return;
}

static int next_dialog = 1;
static int glrKeyAction_Q(glRenderer rnd, int mask)
{
  int   dialog;
  glrGetDialog(rnd, &dialog);
  glrSetDialog(rnd, !dialog);
  next_dialog = 0;
  return GL_TRUE;
}

static int glrKeyAction_q(glRenderer rnd, int mask)
{
  int   dialog;
  glrGetDialog(rnd, &dialog);
  glrSetDialog(rnd, !dialog);
  next_dialog = 1;
  return GL_TRUE;
}

static int glrKeyAction_s(glRenderer rnd, int mask)
{
  glrSetDialog(rnd, 1);
  next_dialog = 1;
  return GL_TRUE;
}

static int glrKeyAction_X(glRenderer rnd, int mask)
{
  exit(0);

  return GL_TRUE;
}

GLTOOLS_WINDOW open_gltools_window(const char *title, const char *geometry,
				   const REAL *world, MESH *mesh, int dialog)
{
  FUNCNAME("open_gltools_window");
  REAL             xmin[3], xmax[3];
  glRenderer       rnd;
  glRendererState  state;
  int              x = 0, y = 0, w = 300, h = 300;
  const char       *s;

  INFO(0,2)("\n");
  if (geometry)
  {
    if (strchr(geometry, 'x'))
    {
      if ((s = strchr(geometry, '+')))
      {
	if (strchr(s+1, '+'))   /*  "wwxhh+xx+yy"   */
	  sscanf(geometry, "%dx%d+%d+%d", &w, &h, &x, &y);
	else                     /*  "wwxhh+xx"      */
	  sscanf(geometry, "%dx%d+%d", &w, &h, &x);
      }
      else                       /*  "wwxhh"         */
      {
	sscanf(geometry, "%dx%d", &w, &h);
      }
    }
    else
    {
      if ((s = strchr(geometry, '+')))
      {
	if (strchr(s+1, '+'))    /*  "xx+yy"         */
	  sscanf(s, "+%d+%d", &x, &y);
	else                     /*  "wwxhh+xx"      */
	  sscanf(s, "+%d", &x);
      }
    }
  }
  if (!title)
    title = "ALBERT";
  rnd = glrCreate((char *) title, x, y, w, h);
  glrSetDialog(rnd, dialog);
  state = glrGetRendererState(rnd);
  state->sensitivity *= M_SQRT2;

  glrRegisterKeyAction(rnd, GLW_Q, glrKeyAction_Q,"Q: quit dialog");
  glrRegisterKeyAction(rnd, GLW_q, glrKeyAction_q,"q: quit");
  glrRegisterKeyAction(rnd, GLW_s, glrKeyAction_s,"s: enter dialog");
  glrRegisterKeyAction(rnd, GLW_X, glrKeyAction_X,"X: exit program");

  if (world)
  {
    xmin[0] = world[0];
    xmax[0] = world[1];
    xmin[1] = world[2];
    xmax[1] = world[3];
#if DIM == 3
    xmin[2] = world[4];
    xmax[2] = world[5];
#endif
  }
  else if (mesh)
  {
    xmin_xmax(mesh, xmin, xmax);
  }
  else
  {
    xmin[0] = 0.0;
    xmax[0] = 1.0;
    xmin[1] = 0.0;
    xmax[1] = 1.0;
#if DIM == 3
    xmin[2] = 0.0;
    xmax[2] = 1.0;
#endif
  }

#if DIM == 2
  glrSetAxisName(rnd, 0, "x");
  glrSetAxisName(rnd, 1, "y");
  glrSetAxisName(rnd, 2, "f(x,y)");
  xmin[2] = -1.0;
  xmax[2] = 1.0;
#elif DIM == 3
  glrSetAxisName(rnd, 0, "x");
  glrSetAxisName(rnd, 1, "y");
  glrSetAxisName(rnd, 2, "z");
#endif
  glrSetVolume(rnd, xmin[0], xmax[0], xmin[1], xmax[1], xmin[2], xmax[2]);

  return((GLTOOLS_WINDOW)  rnd);
}

void close_gltools_window(GLTOOLS_WINDOW win)
{
  if (win)
    glrDestroy((glRenderer) win);
  return;
}

void gltools_drv(GLTOOLS_WINDOW win, const DOF_REAL_VEC *u,
		 REAL min, REAL max)
{
  FUNCNAME("gltools_drv");
  static GLTOOLS_INFO *first_gltools_info = nil;
  GLTOOLS_INFO        *info = nil;
  MESH                *mesh;
  int                 degree;
  int                 ndof;
  glMesh              glm;
  glRenderer          rnd;

  INFO(0,2)("\n");
  if (!u || !win)  return;

  if (!first_gltools_info)  /*  first time, initialize gl_coords */
  {
    int i;
    for (i = 0; i < N_VERTICES; i++)
      gl_coords[i] = nil;
    for (; i < GL_MAX_COORD; i++)
      gl_coords[i] = p[i-N_VERTICES];
  }
  if (!strstr(u->fe_space->bas_fcts->name, "lagrange") &&
    !strstr(u->fe_space->bas_fcts->name, "_disc"))
  {
    MSG("name = %s\n", u->fe_space->bas_fcts->name);
    MSG("only for continuous and discontinuous Lagrange elements\n");
    return;
  }

  rnd = (glRenderer) win;
  glrGetDialog(rnd, &next_dialog);

  mesh       = u->fe_space->mesh;
  degree     = u->fe_space->bas_fcts->degree;
  ndof       = u->fe_space->admin->size_used;

  FOR_ALL_FREE_DOFS(u->fe_space->admin, if (dof < ndof) u->vec[dof] = 0.0);

  if (degree > 4)
  {
    MSG("only for degree <= 4\n");
    return;
  }

  for (info = first_gltools_info; info; info = info->next)
    if (info->data == u) break;

  if (!info)
  {
    info = MEM_ALLOC(1, GLTOOLS_INFO);
    info->next = first_gltools_info;
    first_gltools_info = info;

    info->degree      = degree;
    info->refinement  = 0;
    info->data        = (void *) u;
    info->gl_elements = 1;
    info->gl_coords   = N_VERTICES;
    info->nc          = nc1;
    info->lagrange    = nil;
  }

  glm = glmCreate(ndof, mesh->n_elements, DIM, info, m_cell_loop);
  glrRegisterKeyAction(rnd, GLW_P, glrKeyAction_P,"P: use Lagrange grid");
  key_P_info = info;

  if (u->name)
    glrSetUserInfo(rnd, "%s", u->name);

  glmSetVoffset(glm, 0);
  glrSetInfoCallback(rnd, (glrDrawCallback) glmDrawInfo);

  if (ABS(min - max) < 1.e-25) min++;
  glmSetFunction(glm, u->vec, min, max);
  glRender(rnd, (glrDrawCallback)(glmDraw), glm);
  glmDestroy(glm);

  key_P_info = nil;
  glrRegisterKeyAction(rnd, GLW_P, nil, nil);
  glrSetDialog(rnd, next_dialog);

  return;
}

void gltools_drv_d(GLTOOLS_WINDOW win, const DOF_REAL_D_VEC *u, 
		   REAL min, REAL max)
{
  FUNCNAME("gltools_drv_d");
  static DOF_REAL_VEC uabs = {nil};
  char         name[128];

  INFO(0,2)("\n");

  if (!win || !u)  return;

  sprintf(name, "|%s|", NAME(u));
  uabs.fe_space = u->fe_space;
  uabs.size     = u->fe_space->admin->size_used;
  uabs.vec      = MEM_ALLOC(uabs.size, REAL);
  uabs.name     = name;

  FOR_ALL_DOFS(u->fe_space->admin, 
	       uabs.vec[dof] = NORM_DOW(u->vec[dof]));
  gltools_drv(win, &uabs, min, max);
  MEM_FREE(uabs.vec, uabs.size, REAL);

  return;
}

static void est_loop(glMesh m, void *data, glmSimplexCallback sxf)
{
  MESH             *mesh = (MESH *) data;
  int              i, ic = 0;
  EL_INFO          *el_info;
  REAL             *coord[4];
  TRAVERSE_STACK   *stack = get_traverse_stack();
  int              dof[N_VERTICES];

  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS);
  do
  {
    for (i = 0; i < N_VERTICES; i++)
    {
      coord[i] = el_info->coord[i];
      dof[i] = ic;
    }
    sxf(m, ic+1, 1, NULL, dof, coord);
    ic++;
  } while ((el_info = traverse_next(stack,el_info)));

  free_traverse_stack(stack);
}

void gltools_est(GLTOOLS_WINDOW win, MESH *mesh, REAL (*get_est)(EL *),
		 REAL min, REAL max)
{
  FUNCNAME("gltools_est");
  glMesh          glm;
  glRenderer      rnd;
  int             n;
  REAL            *est_vec = nil;
  EL_INFO          *el_info;
  TRAVERSE_STACK   *stack;

  INFO(0,2)("\n");

  if (!mesh || !win || !get_est)  return;

  rnd = (glRenderer) win;
  glrGetDialog(rnd, &next_dialog);

  est_vec  = MEM_ALLOC(mesh->n_elements, REAL);
  stack = get_traverse_stack();
  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL);
  n = 0;
  while (el_info)
  {
    REAL  est = (*get_est)(el_info->el);
    est_vec[n++] = est;
    el_info = traverse_next(stack,el_info);
  }
  free_traverse_stack(stack);

  glm=glmCreate(mesh->n_elements, mesh->n_elements, DIM, mesh, est_loop);
  glmSetVoffset(glm,0);
  glrSetInfoCallback(rnd,(glrDrawCallback)glmDrawInfo);
  if (ABS(min - max) < 1.e-25) min++;
  glmSetFunction(glm, est_vec, min, max);
  glRender(rnd, (glrDrawCallback)(glmDraw), glm);
  glmDestroy(glm);
  glrSetDialog(rnd, next_dialog);

  MEM_FREE(est_vec, mesh->n_elements, REAL);
  return;
}

struct mesh_loop_data
{
  MESH  *mesh;
  int   mark;
};

static void mesh_loop(glMesh m, void *ud, glmSimplexCallback sxf)
{
  struct mesh_loop_data *data = ud;
  int              i, ic = 1;
  EL_INFO          *el_info;
  REAL             *coord[4];
  TRAVERSE_STACK   *stack = get_traverse_stack();
  int              dof[N_VERTICES];
  
  for(i = 0; i < N_VERTICES; i++)
    dof[i] = 0;

  el_info = traverse_first(stack, data->mesh, -1, CALL_LEAF_EL|FILL_COORDS);
  do
  {
    if (data->mark)
    {
      if (el_info->el->mark > 0)
	for(i = 0; i < N_VERTICES; i++) dof[i] = 1;
      else if (el_info->el->mark < 0)
	for(i = 0; i < N_VERTICES; i++) dof[i] = 2;
      else
	for(i = 0; i < N_VERTICES; i++) dof[i] = 0;
    }

    for(i = 0; i < N_VERTICES; i++)
      coord[i] = el_info->coord[i];
    sxf(m, ic++, 1, NULL, dof, coord);
  } while ((el_info = traverse_next(stack,el_info)));

  free_traverse_stack(stack);
}

void gltools_mesh(GLTOOLS_WINDOW win, MESH *mesh, int mark)
{
  FUNCNAME("gltools_mesh");
  struct mesh_loop_data data[1];
  glMesh          glm;
  glRenderer      rnd;
  double          vec[3] = {0.0, 1.0, -1.0};

  INFO(0,2)("\n");

  if (!(data->mesh = mesh) || !win)  return;

  rnd = (glRenderer) win;
  glrGetDialog(rnd, &next_dialog);

  if ((data->mark = mark))
  {
    glm=glmCreate(3, mesh->n_elements, DIM, data, mesh_loop);
    glmSetVoffset(glm,0);
    glrSetInfoCallback(rnd,(glrDrawCallback)glmDrawInfo);
    glmSetFunction(glm, vec, -1.0, 1.0);
  }
  else
  {
    glm=glmCreate(1, mesh->n_elements, DIM, data, mesh_loop);
    glmSetVoffset(glm,0);
    glrSetInfoCallback(rnd,(glrDrawCallback)glmDrawInfo);
    glmSetFunction(glm, vec, 0.0, 0.0);
  }
  glRender(rnd, (glrDrawCallback)(glmDraw), glm);
  glmDestroy(glm);
  glrSetDialog(rnd, next_dialog);

  return;
}

static void xmin_xmax(MESH *mesh, REAL_D xmin, REAL_D xmax)
{
  MACRO_EL *mel;
  int      n, i;

  if (!mesh) return;
  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    xmin[n] = LARGE;
    xmax[n] = -LARGE;
  }

  for (mel = mesh->first_macro_el; mel; mel = mel->next)
  {
    for (i = 0; i < N_VERTICES; i++)
    {
      if (mel->bound[i])
      {
	for (n = 0; n < DIM_OF_WORLD; n++)
	{
	  xmin[n] = MIN(xmin[n],mel->coord[i][n]);
	  xmax[n] = MAX(xmax[n],mel->coord[i][n]);
	}
      }
    }
  }
  return;
}
