#ifndef _C_OEM_H_
#define _C_OEM_H_

#include <albert_util.h>

/*--------------------------------------------------------------------------*/
/*  library of Orthogonal Error Methods                                     */
/*                                                                          */
/*  most routines are the C-version of OFM-lib by Willy Doerfler            */
/*     http://www.mathematik.uni-kl.de/~doerfler/ofmlib.ps                  */
/*                                                                          */
/*  author: Kunibert G. Siebert                                             */
/*          Institut f"ur Angewandte Mathematik                             */
/*          Albert-Ludwigs-Universit"at Freiburg                            */
/*          Hermann-Herder-Str. 10                                          */
/*          79104 Freiburg                                                  */
/*          Germany                                                         */
/*                                                                          */
/* email:  kunibert@mathematik.uni-freiburg.de                              */
/*                                                                          */
/* (c) by K.G. Siebert (2000)                                               */
/*                                                                          */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*---  data structures for orhtogonal error methods  -----------------------*/
/*--------------------------------------------------------------------------*/

typedef struct oem_data OEM_DATA;
struct oem_data
{
  int        (*mat_vec)(void *, int, const REAL *, REAL *);
  void       *mat_vec_data;
  int        (*mat_vec_T)(void *, int, const REAL *, REAL *);
  void       *mat_vec_T_data;
  void       (*left_precon)(void *, int, REAL *);
  void       *left_precon_data;
  void       (*right_precon)(void *, int, REAL *);
  void       *right_precon_data;

  WORKSPACE  *ws;

  REAL       tolerance;
  int        restart;
  int        max_iter;
  int        info;

  REAL       initial_residual;
  REAL       residual;
};

/*--------------------------------------------------------------------------*/
/*--  implemented solvers  -------------------------------------------------*/
/*--------------------------------------------------------------------------*/

int oem_bicgstab(OEM_DATA *, int, const REAL *, REAL *);
int oem_cg(OEM_DATA *, int, const REAL *, REAL *);
int oem_gmres(OEM_DATA *, int, const REAL *, REAL *);
int oem_gmres_k(OEM_DATA *, int, const REAL *, REAL *);
int oem_odir(OEM_DATA *, int, const REAL *, REAL *);
int oem_ores(OEM_DATA *, int, const REAL *, REAL *);

/*--------------------------------------------------------------------------*/
/*  function call:                                                          */
/*    oem_?(oem_data, dim, b, x);                                           */
/*                                                                          */
/*    input:  oem_data:  above described data structure                     */
/*            dim:       dimension of the linear system                     */
/*            b:         vector storing the right hand side of the sytem    */
/*            x:         vector stroing the initial guess                   */
/*                                                                          */
/*    output: x:         solution of the linear system                      */
/*            oem_data->residual:  residual of the equation                 */
/*                                                                          */
/*    return value:      number of iterations                               */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*  Solver for linear saddle point problem                                  */
/*--------------------------------------------------------------------------*/

typedef struct oem_sp_data OEM_SP_DATA;
struct oem_sp_data
{
  int        (*solve_Auf)(void *, int, const REAL *, REAL *);
  void       *solve_Auf_data;  /*--  solve A u = f  in X  ------------------*/
  void       (*B)(void *, REAL, int, const REAL *, int, REAL *);
  void       *B_data;          /*--  compute B p ---------------------------*/
  void       (*Bt)(void *, REAL, int, const REAL *, int, REAL *);
  void       *Bt_data;         /*--  compute B^* u -------------------------*/
  int        (*project)(void *, int, const REAL *, REAL *);
  void       *project_data;    /*--  project B^* u to Y  -------------------*/
  int        (*precon)(void *, int, const REAL *, const REAL *, REAL *);
  void       *precon_data;     /*--  compute C r in Y  ---------------------*/

  WORKSPACE  *ws;              /*--  workspace  ----------------------------*/

  REAL       tolerance;
  int        restart;
  int        max_iter;
  int        info;

  REAL       initial_residual;
  REAL       residual;
};

/*--------------------------------------------------------------------------*/
/*--  implemented solvers  -------------------------------------------------*/
/*--------------------------------------------------------------------------*/

int oem_spcg(OEM_SP_DATA *, int, const REAL *, REAL *, int, const REAL *,
	     REAL*);
#endif  /* #ifndef _C_OEM_H_ */
