#ifndef _NLS_H_
#define _NLS_H_

#include <albert_util.h>

/*--------------------------------------------------------------------------*/
/*  library of solvers for nonlinear systems                                */
/*                                                                          */
/*  some routines are the C-version of NGL-lib by Willy Doerfler            */
/*     http://www.mathematik.uni-kl.de/~doerfler                            */
/*                                                                          */
/*  author: Kunibert G. Siebert                                             */
/*          Institut f"ur Angewandte Mathematik                             */
/*          Albert-Ludwigs-Universit"at Freiburg                            */
/*          Hermann-Herder-Str. 10                                          */
/*          79104 Freiburg                                                  */
/*          Germany                                                         */
/*                                                                          */
/* email:  kunibert@mathematik.uni-freiburg.de                              */
/*                                                                          */
/* (c) by K.G. Siebert (2000)                                               */
/*                                                                          */
/*--------------------------------------------------------------------------*/

typedef struct nls_data NLS_DATA;
struct nls_data
{
  void       (*update)(void *, int, const REAL *, int, REAL *);
  void       *update_data;
  int        (*solve)(void *, int, const REAL *, REAL *);
  void       *solve_data;
  REAL       (*norm)(void *, int, const REAL *);
  void       *norm_data;

  WORKSPACE  *ws;

  REAL       tolerance;
  int        restart;
  int        max_iter;
  int        info;

  REAL       initial_residual;
  REAL       residual;
};

/*--------------------------------------------------------------------------*/
/*  functions:                                                              */
/*--------------------------------------------------------------------------*/
/*    update:  calculate F(x) and DF(x)                                     */
/*    call:                                                                 */
/*             update(update_data, dim, x, ud_DF, Fx)                       */
/*    arguments:                                                            */
/*             update_data: pointer to user data                            */
/*             dim:         dimension of the nonlinear system               */
/*             x:           vector stroring the actual iterate              */
/*             ud_DF:       ud_DF==1: update DF(x)                          */
/*             Fx:          Fx!=nil:  update F(x) and store it in Fx        */
/*--------------------------------------------------------------------------*/
/*    solve:   solve the linearized system DF(x) d = b for the correction   */
/*    call:                                                                 */
/*             solve(solve_data, dim, b, d)                                 */
/*             solve_data:  pointer to user data                            */
/*             dim:         dimension of the nonlinear system               */
/*             x:           vector stroring the actual iterate              */
/*             b:           right hand side of the linearized system        */
/*             d:           initial guess and solution of the system        */
/*                                                                          */
/*             return value is the number of iterations used by an          */
/*             iterative solver                                             */
/*--------------------------------------------------------------------------*/
/*    norm:    compute the norm of the residual                             */
/*             if norm==nil, the Eucledian norm of the coefficient vector   */
/*             in R^dim is used.                                            */
/*    call:                                                                 */
/*             norm(norm_data, dim, d)                                      */
/*             norm_data:   pointer to user data                            */
/*             dim:         dimension of the nonlinear system               */
/*             d:           coefficient vector of the residual              */
/*                                                                          */
/*             return value is the norm of the residual                     */
/*--------------------------------------------------------------------------*/
/*  ws: workspace                                                           */
/*--------------------------------------------------------------------------*/

int nls_newton(NLS_DATA *ninfo, int dim, REAL *x);
int nls_newton_ds(NLS_DATA *ninfo, int dim, REAL *x);
int nls_newton_fs(NLS_DATA *ninfo, int dim, REAL *x);

int nls_banach(NLS_DATA *ninfo, int dim, REAL *x);
#endif /* #ifndef _NLS_H_ */
