#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>

#define MEM_ALLOC(n,type)      (type *) malloc((n)*sizeof(type))
#define MEM_CALLOC(n,type)      (type *) calloc(n,sizeof(type))
#define MEM_REALLOC(p,nn,type)\
     (type *) p? realloc(p,nn*sizeof(type)) : malloc((nn)*sizeof(type))
#define MEM_FREE(p)  free(p)

#define nil NULL
#define ST(s)           (s ? s : "")
#define IS_BLANK_CHAR(c)   (c == ' ' || c == '\t'|| c == '\f'|| c=='\r')

static const char *funcName;
static FILE  *msgout = nil;
static char  last_char = 0;

void print_msg(const char *format, ...)
{
  static char  funcname[128];
  va_list            arg;

  if (!msgout)  msgout = stdout;

  if (!last_char)
  {
    int   i;
    fprintf(msgout, "%s: ", funcName);
    for (i = 0; i < strlen(funcName) + 2; i++)
      funcname[i] = ' ';
    funcname[i] = '\0';
  }
  else if (last_char == '\n')
    fprintf(stdout, funcname);

  va_start(arg, format);
  vfprintf(msgout, format, arg);
  va_end(arg);
  for (; *format; format++);
  last_char = *--format;
}

#define MSG  print_msg
#define WAIT {char line[10]; MSG("wait for <enter> ..."); gets(line); last_char = '\n';}

/****************************************************************************/

#define COMMENT   '#'
#define DEF_CHAR  '='
#define N_CHAR    1000
#define LENGTH    256

typedef struct parameter PARAMETER;
struct parameter
{
  char  *name;
  char  *value;
  char  *default_value;
  char  *comment;
  int   defined;
};

typedef struct ofiles OFILES;
struct ofiles
{
  char  *s_filename;
  char  *o_filename;

  char  *dependencies;
  char  *include_path;
  char  *vpath;

  OFILES *next;
};

typedef struct makefile_info MAKEFILE_INFO;
struct makefile_info
{
  const char *name;
  const char *makefile_name;

  const char *copy_header;
  const char *source_path;
  const char *lib_path;
  const char *lib_name;
  const char *c_comp;
  const char *cflags;
  const char *f_comp;
  const char *fflags;
  const char *oflags;
  const char *machine_oflag;
  const char *vpath;

  int        def_flags;
  
  OFILES *ofiles;
};

#define N_PARAM   45

struct parameter parameter[N_PARAM] = {
  {"ARCH", nil, nil, "architecture of machine: DEC, LINUX, HP, SGI, SUN"},
  {"MAKE", nil, "make", "make command"},
  {"CC", nil, "cc", "C compiler"},
  {"FC", nil, "f77", "FORTRAN compiler"},
  {"LD", nil, "$(CC)", "linker for ALBERT programs"},
  {"CFLAGS", nil, nil, "compile C flags, -xansi (SGI), e.g."},
  {"FFLAGS", nil, nil, "compile FORTRAN flags"},
  {"OFLAGS", nil, "-O", "optimization flags"},
  {"MACHINE_OFLAG", nil, nil, 
   "machine depended oflag, -mips[1-4] on SGI, e.g."},
  {"DEFAULT_MACHINE_OFLAG", nil, nil, 
   "default machine depended oflag, if several are used"},
  {"RANLIB", nil, nil, "command for ranlib on SUNs, e.g."},
  {"INSTALL_PATH", nil, "$(PWD)", "installation path"},
  {"X11_LIB", nil, "-lX11", "library for X11 graphics"},
  {"USE_OPENGL", nil, "1", 
   "0: no OpenGl; 1: OpenGl in hardware; 2: Mesa OpenGl"},
  {"OPENGL_INCLUDE_PATH", nil, nil, "include path for GL/*.h"},
  {"OPENGL_LIB_PATH", nil, nil, "path of OpenGl lib"},
  {"OPENGL_LIB", nil, "-lGL", "name of OpenGl lib; for Mesa -lMesaGL"},
  {"CREATE_ALBERT_LIB", nil, "1", "0/1: create ALBERT lib"},
  {"ALBERT_LIB_PATH", nil, "$(INSTALL_PATH)/lib", 
   "path for creating ALBERT lib"},
  {"ALBERT_SOURCE_PATH", nil, "$(INSTALL_PATH)/ALBERT/src",
   "path of ALBERT sources"},
  {"DIM", nil, "1 2 3", "ALBERT lib for DIM 1, 2, 3, 1 2, 1 3, 2 3, 1 2 3"},
  {"DIM_OF_WORLD_1D", nil, "1", "DIM_OF_WORLD for 2d lib: 1 2 ..."},
  {"DIM_OF_WORLD_2D", nil, "2", "DIM_OF_WORLD for 2d lib: 2 3 ..."},
  {"DIM_OF_WORLD_3D", nil, "3", "DIM_OF_WORLD for 3d lib: 3 4 ..."},
  {"NEIGH_IN_EL", "0", "0", "library only supports NEIGH_IN_EL=0"},
  {"EL_INDEX", nil, "0", "0, 1, 0 1:store element indices on elements"},
  {"DEBUG_LIB", nil, "0", "debug lib for ALBERT: 0, 1, 0 1"},
  {"DEBUG_FLAGS", nil, "-g", "optimization flag for ALBERT debug lib"},
  {"CPP", nil, "/usr/lib/cpp", "command for C pre processor"},
  {"CREATE_GLTOOLS_INTERFACE", nil, "0", "0/1: install gltools interface"},
  {"GLTOOLS_INCLUDE_PATH", nil, nil, "path of gltools header files"},
  {"GLTOOLS_LIB_PATH", nil, nil, "path of gltools library"},
  {"GLTOOLS_LIB", nil, "-lgltools", "name of gltools lib"},
#if 0
  {"CREATE_GRAPE_INTERFACE", nil, "0", "0/1: install GRAPE interface"},
  {"GRAPE_INCLUDE_PATH", nil, nil, "path of grape.h"},
  {"GRAPE_LIB_PATH", nil, nil, "path of GRAPE library"},
  {"GRAPE_LIB", nil, "-lgr", "name of GRAPE lib"},
#endif
  {"SOLVER_LIB_PATH", nil, nil, "path of solver library"},
  {"SOLVER_LIB", nil, "-lcsolver", "name of existing C-solver library"},
  {"CREATE_SOLVER_LIB", nil, "1", "0/1: install C-solver library"},
  {"SOLVER_SOURCE_PATH", nil, "$(INSTALL_PATH)/SOLVER/src", 
   "path of C-solver sources"},
  {"PLOT_LIB_PATH", nil, nil, "path of existing plot library"},
  {"PLOT_LIB", nil, "-lPLOTansi", "name of existing plot library"},
  {"CREATE_PLOT_LIB", nil, "1", "0/1: install PLOT lib"},
  {"PLOT_SOURCE_PATH", nil, "$(INSTALL_PATH)/PLOT_ANSI/src", 
   "path of PLOT sources"},
  {"BLAS_LIB_PATH", nil, nil, "path of existing BLAS library"},
  {"BLAS_LIB", nil, "-lblas", "name of existing BLAS library"},
  {"CREATE_BLAS_LIB", nil, "0", "0/1: install BLAS1 lib"},
  {"BLAS_SOURCE_PATH", nil, "$(INSTALL_PATH)/BLAS", "path of BLAS1 sources"}
};

/****************************************************************************/
/* print list of all parameters to file fp                                  */
/*   defaults > 0: print default values                                     */
/****************************************************************************/

void print_parameters(FILE *fp, int defaults, int comments) 
{
  int  i;
  
  if (!fp) return;

  for (i = 0; i < N_PARAM; i++)
  {
    if (comments)
    {
      fputc('#', fp);
      fputs(parameter[i].name, fp);
      fputs(": ", fp);
      fputc(' ', fp);
      fputs(ST(parameter[i].comment), fp);
      fputs("; ", fp);
      fputs("default: ", fp);
      if (parameter[i].default_value)
	fputs(ST(parameter[i].default_value), fp);
      else
	fputs("\"\"", fp);
      fputc('\n', fp);
    }
    fputs(parameter[i].name, fp);
    fputs(" = ", fp);
    if (parameter[i].value)
      fputs(parameter[i].value, fp);
    else if (defaults)
      fputs(ST(parameter[i].default_value), fp);
    fputs("\n", fp);
  }
}

#define GET_PARAMETER_VALUE(name)  (get_parameter(name))->value

PARAMETER *get_parameter(const char *name)
{
  int  i;

  for (i = 0; i < N_PARAM; i++)
    if (!strcmp(parameter[i].name, name))  return(parameter+i);
  return(nil);
}

const char *get_parameter_value(const char *name)
{
  PARAMETER *par = get_parameter(name);
  if (par)
  {
    return(par->value);
  }
  else
    return(nil);
}

/****************************************************************************/
static int test(const char *command, const char *file)
{
  char  line[1024];

  if (!file) return(1);
  sprintf(line, "%s %s > /dev/null", ST(command), ST(file));
  return(system(line));
}

static const char *get_word(const char *in, char *word)
{
  if (!in)  return(nil);
  while (*in && isspace(*in)) in++;
  if (!*in)  return(nil);
  while (*in && !isspace(*in)) *word++ = *in++;
  *word = 0;
  return(in);
}

static int fill_vector(const char *s, int *vec, int n)
{
  int   i;
  char  word[10] = "Lall";

  for (i = 0; i < n;)
  {
    if (s = get_word(s, word))
      vec[i++] = atoi(word);
    else
      break;
  }
  return(i);
}

static const char *expand_libname(const char *lib)
{
  static char  name[128];

  if (!lib) return(nil);
  if (lib != strstr(lib, "-l"))  return(nil);
  
  sprintf(name, "lib%s.a", lib+2);
  return(name);
}

static char *gen_lib_name(const char *name, const char *path, 
			  const char *def_flag, const char *mflag)
{
  static char libname[1024];

  if (path)
    sprintf(libname, "%s/lib%s%s%s.a", path, name+2, ST(def_flag), ST(mflag));
  else
    sprintf(libname, "lib%s%s%s.a", name+2, ST(def_flag), ST(mflag));
  return(libname);
}

static int check_error = 0;
static int check_warn = 0;

static void check_parameter(PARAMETER *p)
{
  char       line[1024];
  char   *link_warn = "You may have to change load line for ALBERT programs";
  const char *path, *lib;
  PARAMETER  *par;

  if (!strcmp("ARCH", p->name))
  {
    if (!p->value)
    {
      MSG("ERROR: you must specify %s\n", p->comment);
      check_error++;
    }
  }
  else if (!strcmp("INSTALL_PATH", p->name))
  {
    if (test("test -d", p->value))
    {
      MSG("ERROR: INSTALL_PATH = %s not found;\n", ST(p->value));
      check_error++;
    }
  }
  else if (!strcmp("USE_OPENGL", p->name))
  {
    if (atoi(p->value))
    {
      if (path = get_parameter_value("OPENGL_INCLUDE_PATH"))
      {
	sprintf(line, "%s/GL/glx.h", path);
	if (test("test -f", line))
	{
	  MSG("Warning: Include file GL/glx.h for OpenGl not found in\n");
	  MSG("         OPENGL_INCLUDE_PATH = %s/GL/\n", path);
	  MSG("         Installation without OpenGl graphics\n");
	  p->value = strdup("0");
	  par = get_parameter("OPENGL_LIB_PATH");
	  par->value = nil;
	  check_warn++;
	}
	sprintf(line, "%s/GL/gl.h", path);
	if (test("test -f", line))
	{
	  MSG("Warning: Include file GL/gl.h for OpenGl not found in\n");
	  MSG("         OPENGL_INCLUDE_PATH = %s/GL/\n", path);
	  MSG("         Installation without OpenGl graphics\n");
	  p->value = strdup("0");
	  par = get_parameter("OPENGL_LIB_PATH");
	  par->value = nil;
	  check_warn++;
	}
      }
    }
    if (atoi(p->value))
    {
      if (path = get_parameter_value("OPENGL_LIB_PATH"))
      {
	par = get_parameter("OPENGL_LIB");
	sprintf(line, "%s/%s", path, lib = expand_libname(par->value));
	if (test("test -f", line))
	{
	  MSG("Warning: OpenGl lib %s not found in\n", lib);
	  MSG("         OPENGL_LIB_PATH = %s\n", path);
	  MSG("         %s\n", link_warn);
	  check_warn++;
	  par = get_parameter("OPENGL_LIB_PATH");
	  par->value = nil;
	}
      }
    }
  }
  else if (!strcmp("CREATE_ALBERT_LIB", p->name))
  {
    if (atoi(p->value))
    {
      int i, n_dim, dim[3];
      par = get_parameter("ALBERT_LIB_PATH");
      
      if (test("test -d", par->value))
      {
	MSG("Warning: ALBERT_LIB_PATH = %s not found;\n", ST(par->value));
	MSG("         trying to create it ...\n");
	if (test("mkdir", par->value))
	{
	  MSG("ERROR:    could not create %s\n", ST(par->value));
	  check_error++;
	}
      }
      par = get_parameter("ALBERT_SOURCE_PATH");

      if (test("test -d", par->value))
      {
	MSG("ERROR: ALBERT_SOURCE_PATH = %s not found;\n", ST(par->value));
	check_error++;
      }

      par = get_parameter("DIM");
      n_dim = fill_vector(par->value, dim, 3);
    
      if (n_dim > 3)
      {
	MSG("Warning: ALBERT lib only for DIM = ");
	print_msg("1, 2, 3, 1 2, 1 3, 2 3, 1 2 3\n");
	MSG("         changing value for DIM to 1 2 3\n");
	strcpy(par->value, "1 2 3");
	n_dim = 3;
	check_warn++;
      }
      else if (n_dim == 0)
      {
	MSG("Warning: no dimension DIM specified; using 1 2 3\n");
	par->value = strdup("1 2 3");
	n_dim = 3;
	check_error++;
      }

      for (i = 0; i < n_dim; i++)
      {
	if (dim[i] < 1 || dim[i] > 3)
	{
	  MSG("ERROR: ALBERT lib only for DIM = ");
	  print_msg("1, 2, 3, 1 2, 1 3, 2 3, 1 2 3\n");
	  MSG("         not for DIM = %d\n", dim[i]);
	  n_dim = 0;
	  check_error++;
	}
      }

      if (n_dim > 0)
      {
	n_dim = fill_vector(par->value, dim, 3);

	for (i = 0; i < n_dim; i++)
	{
	  char dow_name[20];
	  int  j, k, n_dow, dow[5];

	  sprintf(dow_name, "DIM_OF_WORLD_%1dD", dim[i]);
	  par = get_parameter(dow_name);
	  n_dow = fill_vector(par->value, dow, 5);
	  for (j = 0; j < n_dow; j++)
	  {
	    if (dow[j] < dim[i])
	    {
	      MSG("Warning: %s must be greater equal %d, %d not allowed\n",
		  dow_name, dim[i], dow[j]);
	      n_dow--;
	      for (k = j; k < n_dow; k++)
		dow[k] = dow[k+1];
	      check_error++;  
	    }
	  }
	  for (j = 0; j < n_dow; j++)
	    sprintf(par->value+(2*j), "%d ", dow[j]);
	}
      }
    }

    par = get_parameter("CPP");

    if (test("ls", par->value))
    {
      MSG("Warning: CPP = %s not found; installation withour CPP\n", p->value);
      par->value = nil;
    }

    par = get_parameter("EL_INDEX");
    if (strcmp(par->value, "0")  && strcmp(par->value, "1") 
	&& strcmp(par->value, "0 1"))
    {
      MSG("Warning: EL_INDEX only for 0, 1, 0 1, not for %s\n", par->value);
      MSG("         using EL_INDEX = 0\n");
      par->value = strdup("0");
      check_warn++;  
    }

    par = get_parameter("DEBUG_LIB");
    if (strcmp(par->value, "0")  && strcmp(par->value, "1") 
	&& strcmp(par->value, "0 1"))
    {
      MSG("Warning: DEBUG_LIB only for 0, 1, 0 1, not for %s\n", par->value);
      MSG("         using DEBUG_LIB = 0\n");
      par->value = strdup("0");
      check_warn++;  
    }
  }
  else if (!strcmp("CREATE_GLTOOLS_INTERFACE", p->name))
  {
    if (!atoi(get_parameter_value("USE_OPENGL")) && atoi(p->value))
    {
      MSG("Warning: no OpenGl graphics\n");
      MSG("         Installation without gltools interface\n");
      p->value = strdup("0");
      check_warn++;
    }

    if (atoi(p->value))
    {
      if (path = (get_parameter("GLTOOLS_INCLUDE_PATH"))->value)
      {
	sprintf(line, "%s/glmesh.h", path);

	if (test("test -f", line))
	{
	  MSG("Warning: Include file glmesh.h for gltools not found in\n");
	  MSG("         GLTOOLS_INCLUDE_PATH = %s\n", path);
	  MSG("         Installation without gltools interface\n");
	  p->value = strdup("0");
	  check_warn++;
	}
	sprintf(line, "%s/glrnd.h", path);

	if (test("test -f", line))
	{
	  MSG("Warning: Include file glrnd.h for gltools not found in\n");
	  MSG("         GLTOOLS_INCLUDE_PATH = %s\n", path);
	  MSG("         Installation without gltools interface\n");
	  p->value = strdup("0");
	  check_warn++;
	}
      }
      else
      {
	MSG("Warning: No GLTOOLS_INCLUDE_PATH specified\n");
	MSG("         Installation without gltools interface\n");
	p->value = strdup("0");
	check_warn++;
      }
    }

    if (atoi(p->value))
    {
      if (path = (get_parameter("GLTOOLS_LIB_PATH"))->value)
      {
	const char  *lib;
	par = get_parameter("GLTOOLS_LIB");

	sprintf(line, "%s/%s", path, lib = expand_libname(par->value));
	if (test("test -f", line))
	{
	  MSG("Warning: gltools lib %s not found in\n", lib);
	  MSG("         GLTOOLS_LIB_PATH = %s\n", ST(path));
	  MSG("         %s\n", link_warn);
	  check_warn++;
	}
      }
    }
  }
#if 0
  else if (!strcmp("CREATE_GRAPE_INTERFACE", p->name))
  {
    if (atoi(p->value))
    {
      path  = (get_parameter("GRAPE_INCLUDE_PATH"))->value;
      sprintf(line, "%s/grape.h", ST(path));
      if (test("test -f", line))
      {
	MSG("Warning: Include file grape.h for GRAPE not found in\n");
	MSG("         GRAPE_INCLUDE_PATH = %s\n", ST(path));
	MSG("         Installation without GRAPE interface\n");
	p->value = strdup("0");
	check_warn++;
      }
    }

    if (atoi(p->value))
    {
      if (path = (get_parameter("GRAPE_LIB_PATH"))->value)
      {
	par = get_parameter("GRAPE_LIB");
	sprintf(line, "%s/%s", path, lib = expand_libname(par->value));
	if (test("test -f", line))
	{
	  MSG("Warning: GRAPE lib %s not found in\n", lib);
	  MSG("         GRAPE_LIB_PATH = %s\n", path);
	  MSG("         %s\n", link_warn);
	  check_warn++;
	}
      }
    }
  }
#endif
  else if (!strcmp("CREATE_SOLVER_LIB", p->name))
  {
    if (!atoi(p->value))
    {
      if (path = (get_parameter("SOLVER_LIB_PATH"))->value)
      {
	par = get_parameter("SOLVER_LIB");
	sprintf(line, "%s/%s", path, lib = expand_libname(par->value));
	if (test("test -f", line))
	{
	  MSG("Warning: SOLVER lib %s not found in\n", lib);
	  MSG("         SOLVER_LIB_PATH = %s\n", path);
	  MSG("         %s\n", link_warn);
	  check_warn++;
	}
      }
    }
    else
    {
      par = get_parameter("SOLVER_LIB");
      lib = par->value;
      path = (get_parameter("SOLVER_SOURCE_PATH"))->value;

      if (test("test -d", path))
      {
	MSG("ERROR: SOLVER_SOURCE_PATH = %s not found;\n", ST(path));
	check_error++;
      }

      par = get_parameter("SOLVER_LIB_PATH");
      if (test("test -d", par->value))
      {
	MSG("Warning: SOLVER_LIB_PATH = %s not found for installation;\n", 
	    ST(par->value));
	path = (get_parameter("ALBERT_LIB_PATH"))->value;
	if (test("test -d", path))
	{
	  par->value = strdup(getenv("PWD"));
	  MSG("         install %s into current directory %s\n",
	      lib, par->value);
	}
	else
	{
	  par->value = (char *) path;
	  MSG("         install %s into ALBERT_LIB_PATH %s\n",
	      lib, par->value);
	}
	check_warn++;
      }
    }
  }
  else if (!strcmp("CREATE_PLOT_LIB", p->name))
  {
    if (!atoi(p->value))
    {
      if (path = (get_parameter("PLOT_LIB_PATH"))->value)
      {
	par = get_parameter("PLOT_LIB");
	sprintf(line, "%s/%s", path, lib = expand_libname(par->value));
	if (test("test -f", line))
	{
	  MSG("Warning: PLOT lib %s not found in\n", lib);
	  MSG("         PLOT_LIB_PATH = %s\n", path);
	  MSG("         %s\n", link_warn);
	  check_warn++;
	}
      }
    }
    else
    {
      lib = (get_parameter("PLOT_LIB"))->value;

      par = get_parameter("PLOT_SOURCE_PATH");
      if (test("test -d", par->value))
      {
	MSG("ERROR: PLOT_SOURCE_PATH = %s not found;\n", ST(par->value));
	check_error++;
      }

      par = get_parameter("PLOT_LIB_PATH");
      if (test("test -d", par->value))
      {
	MSG("Warning: PLOT_LIB_PATH = %s not found for installation;\n", 
	    ST(par->value));

	path = (get_parameter("ALBERT_LIB_PATH"))->value;
	if (test("test -d", path))
	{
	  par->value = strdup(getenv("PWD"));
	  MSG("         install %s into current directory %s\n",
	      lib, par->value);
	}
	else
	{
	  par->value = (char *) path;
	  MSG("         install %s into ALBERT_LIB_PATH %s\n",
	      lib, par->value);
	}
	check_warn++;
      }
    }
  }
  else if (!strcmp("CREATE_BLAS_LIB", p->name))
  {
    if (!atoi(p->value))
    {
      par = get_parameter("BLAS_LIB_PATH");
      if (path = par->value)
      {
	par = get_parameter("BLAS_LIB");
	sprintf(line, "%s/%s", path, lib = expand_libname(par->value));
	if (test("test -f", line))
	{
	  MSG("Warning: BLAS lib %s not found in\n", lib);
	  MSG("         BLAS_LIB_PATH = %s\n", path);
	  MSG("You may have to change load line for ALBERT programs\n");
	  MSG("         %s\n", link_warn);
	  check_warn++;
	}
      }
    }
    else
    {
      lib = (get_parameter("BLAS_LIB"))->value;
      path = (get_parameter("BLAS_SOURCE_PATH"))->value;
      if (test("test -d", path))
      {
	MSG("ERROR: BLAS_SOURCE_PATH = %s not found;\n", ST(p->value));
	check_error++;
      }

      par = get_parameter("BLAS_LIB_PATH");
      if (test("test -d", par->value))
      {
	MSG("Warning: BLAS_LIB_PATH = %s not found for installation;\n",
	    ST(par->value));
	
	path = (get_parameter("ALBERT_LIB_PATH"))->value;
	if (test("test -d", path))
	{
	  par->value = strdup(getenv("PWD"));
	  MSG("         install %s into current directory %s\n",
	      lib, par->value);
	}
	else
	{
	  par->value = (char *) path;
	  MSG("         install %s into ALBERT_LIB_PATH %s\n",
	      lib, par->value);
	}
	check_warn++;
      }
    }
  }
}

/****************************************************************************/
/*  read parameters(): initialize the parameters from file                  */
/****************************************************************************/

static void error_msg(const char *message, int *ln, const char *filename)
{
  MSG("ERROR: %s on ", message);
  if (ln[0] == ln[1])
    print_msg("line %d of file %s\n", ln[0], filename);
  else
    print_msg("lines %d to %d of file %s\n", ln[0], ln[1], filename);
  exit(0);
}

static char *get_line(FILE *fp, int *ln, const char *filename)
{
  static char  line[1024];
  char   *lp = line, *lpc;
  int    size = 0;

  if (!fgets(lp, 1023, fp)) return(nil);
  ln[0] = ++ln[1];

  lp = line;
  do
  {
    size += strlen(lp);

    if (lpc = strstr(lp, "\\\n"))                   /* continuation line  */
    {
      ln[1]++;
      lp = lpc;
      continue;
    }
    if (lpc = strstr(line, "\n")) *lpc = '\0';      /* remove newline char */

    if ((lpc = strstr(lp, "#"))  &&  lpc == line) /* comment line, skip    */ 
    {
      size = 0;
      lp = line;
      continue;
    }
    else if (lpc)                       /* comment sign, skip rest of line  */
      *lpc = '\0';

    lpc = lp;
    while(*lpc) lpc++;                       /*  move to end of string      */
    while(IS_BLANK_CHAR(*lpc)) lpc--;        /* remove blanks at the end    */
      *lpc = '\0';

    if (lpc == line)                        /*  empty line, skip  */
    {
      size = 0;
      lp = line;
      continue;
    }

    if (!(lpc = strstr(line, "=")))
      error_msg("missing seperator", ln, filename);
    else if (lpc == line)
      error_msg("missing variable name", ln, filename);
    else
      break;
  }
  while(fgets(lp, 1023 - size, fp));

  for (lpc = lp = line; IS_BLANK_CHAR(*lpc); lpc++);
  if (lpc != lp)                       /* remove leading blank characters  */
    while (*lp++ = *lpc++);

  return(line);
}

int read_parameters(const char *filename)
{
  PARAMETER *par;
  char      *line;
  char      *var, *val;
  int       ln[2] = {0, 0};
  FILE      *fp = fopen(filename, "r");

  if (!fp)
  {
    MSG("can not open %s for reading\n");
    exit(1);
  }

  while (line = get_line(fp, ln, filename))
  {
    for (var = line; *var && !IS_BLANK_CHAR(*var) && *var != '='; var++);
      
    for (val  = var; *val && *val != '='; val++);        /* loop to '='     */
    if (*val == '=') val++;
    for (; *val && IS_BLANK_CHAR(*val);) val++;          /* skip blank char */

    *var = '\0';   /*  make variable name to string, values stored at value */
    var = line;

    if (par = get_parameter(var))                      /*  get parameter    */
    {
      char *c1 = val, *c2 = val;

      while (*c2)
      {
	for (; *c2 && !IS_BLANK_CHAR(*c2); *c1++ = *c2++);
	if (IS_BLANK_CHAR(*c2)) *c1++ = *c2++;           /* one blank char  */
	for (; *c2 && IS_BLANK_CHAR(*c2); c2++);         /* skip blank char */
      }
      *c1 = '\0';
      if (*val)
      {
	if (strstr(par->name, "_PATH"))
	  while (strlen(val) > 1 && *--c1 == '/') *c1 = '\0';
	par->value = strdup(val);
      }
      else
	par->value = nil;
    }
    else
    {
      MSG("no variable %s to be defined; skipping ", var);
      if (ln[0] == ln[1])
	print_msg("line %d of file %s\n", ln[0], filename);
      else
	print_msg("lines %d to %d of file %s\n", ln[0], ln[1], filename);
    }
  }
  fclose(fp);
  return(1);
}

static int init_parameter(struct parameter *p);

static void expand_definition(char *pos, int size) /* size length of pos */
{
  PARAMETER  *par;
  char       name[1024];
  char       var[64], *value, *pvar, *pv;
  char       *dollar1, *dollar2;

  if (dollar1 = strstr(pos, "$("))
  {
    dollar2 = dollar1+2;
    size = size - (dollar1 - pos);
    expand_definition(dollar2, size - 2);
    if (!(dollar2 = strstr(dollar2, ")")))
    {
      MSG("ERROR: $( without )\n");
      exit(0);
    }
    for (pvar = var, pv = dollar1+2; pv < dollar2; *pvar++ = *pv++);
    *pvar = '\0';                /*  get variable name $(...)             */

    if (par = get_parameter(var))             /* make variable            */
    {
      init_parameter(par);
      value = par->value;
    }
    else                                      /* environment variable     */
    {
      value = getenv(var);
    }

    dollar2++;                   /*  produce expanded name                */
    if (value)    size -= strlen(value);
    if (dollar2)  size -= strlen(dollar2);
    if (size < 1)
    {
      MSG("expanded name longer than 1023 characters;\n");
      MSG("can not handle such long names\n");
      exit(1);
    }
    sprintf(name, "%s%s", ST(value), ST(dollar2));
    strcpy(dollar1, name);        /*  store expanded name                  */
  }

  return;
}

static int init_parameter(struct parameter *p)
{
  char  line[1024];

  if (!p->value && !p->default_value)
    return(1);

  if (p->value)
    strcpy(line, p->value);   /*  string long enough to store expansion     */
  else
    strcpy(line, p->default_value);

  expand_definition(line, sizeof(line));

  if (p->value  && (strlen(p->value) >= strlen(line)))
    strcpy(p->value, line);
  else
    p->value = strdup(line);

  return(1);
}

static int init_parameters(const char *filename)
{
  int  i;
  
  if (!read_parameters(filename))
  {
    MSG("ERROR during reading of %s\n", filename);
    return(0);
  }
  for (i = 0; i < N_PARAM; i++)
    init_parameter(parameter+i);

  check_error = check_warn = 0;
  for (i = 0; i < N_PARAM; i++)
    check_parameter(parameter+i);
  if (check_error)
  {
    MSG("%d ERRORs detected during check of variables\n", check_error);
    MSG("Check messages and change variable values\n");
  }
  if (check_warn)
  {
    MSG("%d warnings detected during check of variables\n", check_warn);
    MSG("Check messages and variable values\n");
  }
  return(check_error);
}

OFILES *swap(OFILES *ofile)
{
  OFILES  *next = ofile->next;
  int     n;

  if (!next || (n = strcmp(ofile->o_filename, next->o_filename)) <= 0)
  {
    return(ofile);  /* no swap is needed  */
  }
  else if (!n)      /* function already in list, free memeory   */
  {
    MEM_FREE(ofile);
    return(next);
  }
  else              /*  swap to elements; compare in the remaining list  */
  {
    ofile->next = next->next;
    next->next = swap(ofile);
    return(next);
  }
}

static OFILES *create(OFILES *head, const char *name, const char *depend, 
		      const char *include, const char *vpath)
{
  char   *po;
  OFILES  *file;
  
  file = MEM_CALLOC(1, OFILES);
  file->s_filename = strdup(name);
  file->o_filename = strdup(name);
  for (po = file->o_filename; *(po+1); po++);
  *po = 'o';

  if (depend)  file->dependencies = strdup(depend);
  if (include) file->include_path = strdup(include);
  if (vpath)   file->vpath = strdup(vpath);
  file->next = head;

  return(swap(file));
}

void print_filenames(OFILES *head)
{
  for (; head; head = head->next)
  {
    printf("program name: %s\n",head->s_filename);
    printf("O program name: %s\n",head->o_filename);
    printf("Dependencies:   %s\n",head->dependencies);
    printf("Include Path:   %s\n\n",head->include_path);
  }
}

static void init_albert_files(MAKEFILE_INFO info[4])
{
  char        line[256], line2[256];
  OFILES      *albert = nil;
  const char  *header = "albert.h albert_util.h";
  const char  *inc, *value;
  int         dim;

  for (dim = 1; dim <= 3; dim++)
  {
    albert = nil;
    albert = create(albert, "adapt.c", header, nil, "../Common");
    albert = create(albert, "assemble.c", header, nil, "../Common");
    albert = create(albert, "check.c", header, nil, "../Common");
    albert = create(albert, "dof_admin.c", header, nil, "../Common");
    albert = create(albert, "element.c", header, nil, nil);
    albert = create(albert, "estimator.c", header, nil, "../Common");
    albert = create(albert, "error.c", header, nil, "../Common");
    albert = create(albert, "eval.c", header, nil, "../Common");
    albert = create(albert, "HB_precon.c", header, nil, "../Common");
    albert = create(albert, "l2scp.c", header, nil, "../Common");
    albert = create(albert, "MG.c", header, nil, "../Common");
    albert = create(albert, "MG_s.c", header, nil, "../Common");
    albert = create(albert, "MG_s1.c", header, nil, "../Common");
    albert = create(albert, "MG_s2.c", header, nil, "../Common");
    albert = create(albert, "memory.c", header, nil, "../Common");
    albert = create(albert, "numint.c", header, nil, "../Common");
    albert = create(albert, "read_mesh.c", header, nil, "../Common");
    albert = create(albert, "read_mesh_xdr.c", header, nil, "../Common");
    albert = create(albert, "sor.c", header, nil, "../Common");
    albert = create(albert, "ssor.c", header, nil, "../Common");
    if (dim > 1)
      albert = create(albert, "trav_xy.c", header, nil, "../Common");
    albert = create(albert, "write_mesh.c", header, nil, "../Common");
    albert = create(albert, "write_mesh_xdr.c", header, nil, "../Common");
    if (dim == 2)
      albert = create(albert, "write_mesh_ps.c", header, nil, nil);
  
    if (value = get_parameter_value("CPP"))
    {
      sprintf(line, "-DCPP=\\\"%s\\\"", value);
      albert = create(albert, "util.c", header, line, "../Common");
    }
    else
    {
      albert = create(albert, "util.c", header, nil, "../Common");
    }

    if (dim == 1)
      sprintf(line, "%s %s", header, "lagrange_[0-2].c");
    else
      sprintf(line, "%s %s", header, "lagrange_[0-4].c");
    albert = create(albert, "bas_fct.c", line, nil, "../Common");

    sprintf(line, "%s %s", header, "../Common/coarsen_common.c");
    albert = create(albert, "coarsen.c", line, nil, nil);
    sprintf(line, "%s %s", header, "../Common/macro_common.c");
    albert = create(albert, "macro.c", line, nil, nil);
    if (dim > 1)
    {
      sprintf(line, "%s %s", header, "../Common/level_common.c");
      albert = create(albert, "level.c", line, nil, nil);
    }
    sprintf(line, "%s %s", header, "../Common/refine_common.c");
    albert = create(albert, "refine.c", line, nil, nil);
    sprintf(line, "%s %s", header, "../Common/traverse_r_common.c");
    albert = create(albert, "traverse_r.c", line, nil, nil);
    sprintf(line, "%s %s", header, "../Common/traverse_nr_common.c");
    albert = create(albert, "traverse_nr.c", line, nil, nil);

    sprintf(line, "%s %s/oem.h", header, 
	    get_parameter_value("SOLVER_SOURCE_PATH"));
    sprintf(line2, "-I. -I%s", get_parameter_value("SOLVER_SOURCE_PATH"));

    albert = create(albert, "oem_solve_s.c", line, line2, "../Common");
    albert = create(albert, "oem_solve_d.c", line, line2, "../Common");

    if (dim < 3)
    {
      if (atoi(get_parameter_value("USE_OPENGL")))
      {
	if (inc = get_parameter_value("OPENGL_INCLUDE_PATH"))
	  sprintf(line, "-DUSE_OPENGL=1 -I%s -I", inc);
	else
	  sprintf(line, "-DUSE_OPENGL=1 -I");
	strcat(line, get_parameter_value("ALBERT_LIB_PATH"));
	albert = create(albert, "graphXO.c", header, line, nil);
      }
      else
      {
	sprintf(line, "-DUSE_OPENGL=0 -I%s", 
		get_parameter_value("ALBERT_LIB_PATH"));
	albert = create(albert, "graphXO.c", header, line, nil);
      }
    }
    else
    {
      albert = create(albert, "graph_none.c", nil, nil, nil);
    }

    if (atoi(get_parameter_value("CREATE_GLTOOLS_INTERFACE")))
    {
      char  line2[256];
      inc = get_parameter_value("GLTOOLS_INCLUDE_PATH");
      sprintf(line, "%s/gl*.h", inc);
      sprintf(line2, "-I%s", inc);
      if (inc = get_parameter_value("OPENGL_INCLUDE_PATH"))
      {
	strcat(line2, " -I");
	strcat(line2, inc);
      }
      albert = create(albert, "gltools.c", line, line2, "../Common");
    }

    sprintf(line, "ALBERT-%d", dim);
    info[dim].name          = strdup(line);
    info[dim].copy_header   = strdup(header);
    sprintf(line, "%s/%1dd", get_parameter_value("ALBERT_SOURCE_PATH"), dim);
    info[dim].source_path   = strdup(line);
    info[dim].lib_path      = get_parameter_value("ALBERT_LIB_PATH");
    info[dim].lib_name      = strdup("-lALBERT");
    info[dim].c_comp        = get_parameter_value("CC");
    info[dim].cflags        = get_parameter_value("CFLAGS");
    info[dim].f_comp        = nil;
    info[dim].fflags        = nil;
    info[dim].oflags        = get_parameter_value("OFLAGS");
    info[dim].machine_oflag = get_parameter_value("MACHINE_OFLAG");
    info[dim].vpath         = strdup("../Common");

    info[dim].def_flags     = 1;
    info[dim].ofiles        = albert;

    sprintf(line, "Makefile.%s", get_parameter_value("ARCH"));
    info[dim].makefile_name = strdup(line);
  }
  return;
}

static void init_blas_files(MAKEFILE_INFO *info)
{
  char    line[256];
  OFILES  *blas = nil;
  
  blas = create(blas, "caxpy.f", nil, nil, nil);
  blas = create(blas, "ccopy.f", nil, nil, nil);
  blas = create(blas, "cdotc.f", nil, nil, nil);
  blas = create(blas, "cdotu.f", nil, nil, nil);
  blas = create(blas, "crotg.f", nil, nil, nil);
  blas = create(blas, "cscal.f", nil, nil, nil);
  blas = create(blas, "csrot.f", nil, nil, nil);
  blas = create(blas, "csscal.f", nil, nil, nil);
  blas = create(blas, "cswap.f", nil, nil, nil);
  blas = create(blas, "dasum.f", nil, nil, nil);
  blas = create(blas, "daxpy.f", nil, nil, nil);
  blas = create(blas, "dcabs1.f", nil, nil, nil);
  blas = create(blas, "dcopy.f", nil, nil, nil);
  blas = create(blas, "ddot.f", nil, nil, nil);
  blas = create(blas, "dnrm2.f", nil, nil, nil);
  blas = create(blas, "drot.f", nil, nil, nil);
  blas = create(blas, "drotg.f", nil, nil, nil);
  blas = create(blas, "drotm.f", nil, nil, nil);
  blas = create(blas, "drotmg.f", nil, nil, nil);
  blas = create(blas, "dscal.f", nil, nil, nil);
  blas = create(blas, "dset.f", nil, nil, nil);
  blas = create(blas, "dswap.f", nil, nil, nil);
  blas = create(blas, "dzasum.f", nil, nil, nil);
  blas = create(blas, "dznrm2.f", nil, nil, nil);
  blas = create(blas, "icamax.f", nil, nil, nil);
  blas = create(blas, "idamax.f", nil, nil, nil);
  blas = create(blas, "isamax.f", nil, nil, nil);
  blas = create(blas, "izamax.f", nil, nil, nil);
  blas = create(blas, "sasum.f", nil, nil, nil);
  blas = create(blas, "saxpy.f", nil, nil, nil);
  blas = create(blas, "scasum.f", nil, nil, nil);
  blas = create(blas, "scnrm2.f", nil, nil, nil);
  blas = create(blas, "scopy.f", nil, nil, nil);
  blas = create(blas, "sdot.f", nil, nil, nil);
  blas = create(blas, "snrm2.f", nil, nil, nil);
  blas = create(blas, "srot.f", nil, nil, nil);
  blas = create(blas, "srotg.f", nil, nil, nil);
  blas = create(blas, "srotm.f", nil, nil, nil);
  blas = create(blas, "srotmg.f", nil, nil, nil);
  blas = create(blas, "sscal.f", nil, nil, nil);
  blas = create(blas, "sswap.f", nil, nil, nil);
  blas = create(blas, "zaxpy.f", nil, nil, nil);
  blas = create(blas, "zcopy.f", nil, nil, nil);
  blas = create(blas, "zdotc.f", nil, nil, nil);
  blas = create(blas, "zdotu.f", nil, nil, nil);
  blas = create(blas, "zdrot.f", nil, nil, nil);
  blas = create(blas, "zdscal.f", nil, nil, nil);
  blas = create(blas, "zrotg.f", nil, nil, nil);
  blas = create(blas, "zscal.f", nil, nil, nil);
  blas = create(blas, "zswap.f", nil, nil, nil);

  info->name          = strdup("BLAS1");
  info->copy_header   = nil;
  info->source_path   = get_parameter_value("BLAS_SOURCE_PATH");
  info->lib_path      = get_parameter_value("BLAS_LIB_PATH");
  info->lib_name      = get_parameter_value("BLAS_LIB");
  info->c_comp        = nil;
  info->cflags        = nil;
  info->f_comp        = get_parameter_value("FC");
  info->fflags        = get_parameter_value("FFLAGS");
  info->oflags        = get_parameter_value("OFLAGS");
  info->machine_oflag = get_parameter_value("MACHINE_OFLAG");

  info->ofiles        = blas;
  info->def_flags     = 0;

  sprintf(line, "Makefile.%s", get_parameter_value("ARCH"));
  info->makefile_name = strdup(line);

  return;
}

static void init_csolver_files(MAKEFILE_INFO *info)
{
  char     line[128];
  OFILES   *solver = nil;

  solver = create(solver, "daxpy.c", "cblas.h", nil, nil);
  solver = create(solver, "dcopy.c", "cblas.h", nil, nil);
  solver = create(solver, "ddot.c", "cblas.h", nil, nil);
  solver = create(solver, "dexpy.c", "cblas.h", nil, nil);
  solver = create(solver, "dexpy_.f", nil, nil, nil);
  solver = create(solver, "dmxpy.c", "cblas.h", nil, nil);
  solver = create(solver, "dmxpy_.f", nil, nil, nil);
  solver = create(solver, "dnrm2.c", "cblas.h", nil, nil);
  solver = create(solver, "dscal.c", "cblas.h", nil, nil);
  solver = create(solver, "dset.c", "cblas.h", nil, nil);
  solver = create(solver, "dswap.c", "cblas.h", nil, nil);
  solver = create(solver, "dxpay.c", "cblas.h", nil, nil);
  solver = create(solver, "dxpay_.f", nil, nil, nil);

  sprintf(line, "-I%s/Common", get_parameter_value("ALBERT_SOURCE_PATH"));
  solver = create(solver, "bicgstab.c", "oem.h", line, nil);
  solver = create(solver, "cg.c", "oem.h", line, nil);
  solver = create(solver, "gmres.c", "oem.h", line, nil);
  solver = create(solver, "info.c", "oem.h", line, nil);
  solver = create(solver, "odir.c", "oem.h", line, nil);
  solver = create(solver, "ores.c", "oem.h", line, nil);
  solver = create(solver, "spcg.c", "oem.h", line, nil);

  solver = create(solver, "newton.c", "nls.h", line, nil);
  solver = create(solver, "newton_ds.c", "nls.h", line, nil);
  solver = create(solver, "newton_fs.c", "nls.h", line, nil);
  solver = create(solver, "banach.c", "nls.h", line, nil);

  info->name          = strdup("C-solver");
  info->copy_header   = strdup("oem.h nls.h cblas.h");
  info->source_path   = get_parameter_value("SOLVER_SOURCE_PATH");
  info->lib_path      = get_parameter_value("SOLVER_LIB_PATH");
  info->lib_name      = get_parameter_value("SOLVER_LIB");
  info->c_comp        = get_parameter_value("CC");
  info->cflags        = get_parameter_value("CFLAGS");
  info->f_comp        = get_parameter_value("FC");
  info->fflags        = get_parameter_value("FFLAGS");
  info->oflags        = get_parameter_value("OFLAGS");
  info->machine_oflag = get_parameter_value("MACHINE_OFLAG");

  info->def_flags     = 0;
  info->ofiles        = solver;
  
  sprintf(line, "Makefile.%s", get_parameter_value("ARCH"));
  info->makefile_name = strdup(line);
}

static void init_plot_files(MAKEFILE_INFO *info)
{
  char        line[256];
  OFILES      *plot = nil;
  const char  *arch = get_parameter_value("ARCH");
  const char  *include = "-I/usr/include/X11";
  const char  *header = "PLOTansi.h";

  if (!strcmp(arch, "IBM"))
    plot = create(plot, "PLOT_for_ibm.c", header, include, nil);
  else if (!strcmp(arch, "TITAN"))
    plot = create(plot, "PLOT_for_tit.c", header, include, nil);
  else if (!strcmp(arch, "SUN"))
    plot = create(plot, "PLOT_for_sun.c", header, include, nil);
  else if (!strcmp(arch, "SGI"))
    plot = create(plot, "PLOT_for_sun.c", header, include, nil);
  else if (!strcmp(arch, "LINUX"))
    plot = create(plot, "PLOT_for_sun.c", header, include, nil);
  else if (!strcmp(arch, "DEC"))
    plot = create(plot, "PLOT_for_sun.c", header, include, nil);
  else if (!strcmp(arch, "HP"))
    plot = create(plot, "PLOT_for_sun.c", header, include, nil);
  else
  {
    MSG("ERROR: can not create PLOTansi lib for architecture %s\n", arch);
    return;
  }

  plot = create(plot, "PLOT_open.c", header, include, nil);
  plot = create(plot, "PLOT2d.c", header, include, nil);
  plot = create(plot, "PLOTlog.c", header, include, nil);

  info->name          = strdup("PLOTansi");
  info->copy_header   = strdup(header);
  info->source_path   = get_parameter_value("PLOT_SOURCE_PATH");
  info->lib_path      = get_parameter_value("PLOT_LIB_PATH");
  info->lib_name      = get_parameter_value("PLOT_LIB");
  info->c_comp        = get_parameter_value("CC");
  info->cflags        = get_parameter_value("CFLAGS");
  info->f_comp        = nil;
  info->fflags        = nil;
  info->oflags        = get_parameter_value("OFLAGS");
  info->machine_oflag = get_parameter_value("MACHINE_OFLAG");

  info->def_flags     = 0;
  info->ofiles        = plot;

  sprintf(line, "Makefile.%s", get_parameter_value("ARCH"));
  info->makefile_name = strdup(line);

  return;
}

#include <time.h>

#define NL(fp) fputc('\n', fp)

void print_comment(FILE *mkfi, const char *comment)
{
  if (!mkfi)  return;

  if (comment)
    fprintf(mkfi, "#   %-67s#\n", comment);
  else
  {
    int  i;
    for (i = 0; i < 72; i++) 
      fputc('#', mkfi);
    NL(mkfi);
  }
  return;
}

void print_lib_files(OFILES *ofiles, FILE *fp)
{
  char  name[256];
  int   n_files = 0;

  if (!ofiles)  return;

  fprintf(fp, "LIB_FILES = ");
  for (; ofiles; ofiles = ofiles->next)
  {
    if (n_files++%2)
    {
      if (ofiles->next)
	fprintf(fp, "$(LIB)(%s)\\\n%12s", ofiles->o_filename, "");
      else
	fprintf(fp, "$(LIB)(%s)\n%12s", ofiles->o_filename, "");
    }
    else
    {
      sprintf(name, "$(LIB)(%s)", ofiles->o_filename);
      fprintf(fp, "%-25s", name);
    }
  }
  if (n_files%2) NL(fp);
}

#define UNKNOWN_FILE 0
#define C_FILE       1
#define F_FILE       2

static char source_type(const char *name)
{
  const char  *type;

  for (type = name; *type; type++);
  type--;

  if (*type != 'c'  && *type != 'f' || *(type-1) != '.')
  {
    MSG("only for `*.c' or `.f' files\n");
    return(UNKNOWN_FILE);
  }
  else if (*type == 'c')
    return(C_FILE);
  else if (*type == 'f')
    return(F_FILE);
  else
    return(UNKNOWN_FILE);
}

void print_compile(OFILES *ofiles, FILE *fp)
{
  if (!ofiles)  return;

  for (; ofiles; ofiles = ofiles->next)
  {
    int  type;

    if (type = source_type(ofiles->s_filename))
    {
      fprintf(fp, "$(LIB)(%s): %s", ofiles->o_filename, ofiles->s_filename);
      if (ofiles->dependencies)
	fprintf(fp, " %s", ofiles->dependencies);
      NL(fp);
      if (type == C_FILE)
	fprintf(fp, "\t$(CC) -c $(CFLAGS)");
      else if (type == F_FILE)
	fprintf(fp, "\t$(FC) -c $(FFLAGS)");
      if (ofiles->include_path)
	fprintf(fp, " %s", ofiles->include_path);
      if (ofiles->vpath)
	fprintf(fp, " %s/%s", ofiles->vpath, ofiles->s_filename);
      else
	fprintf(fp, " %s", ofiles->s_filename);
      fprintf(fp, " -o %s", ofiles->o_filename);
      NL(fp);
      fprintf(fp, "\t$(AR) $(ARFLAGS) $(LIB) %s\n", ofiles->o_filename);
      fprintf(fp, "\t-rm %s\n", ofiles->o_filename);
    }
  }
}

static int make_makefile(MAKEFILE_INFO *info)
{
  time_t     act_time;
  FILE       *fp;
  char       line[256];
  const char *lib_name;
  const char *dflags, *moflag, *libext, *ranlib;
  char       *date, *pn;

  if (!info->source_path) info->source_path = ".";
  sprintf(line, "%s/%s", info->source_path, info->makefile_name);

  if (!(fp = fopen(line, "w")))
  {
    MSG("ERROR: can not write Makefile %s\n", info->makefile_name);
    return(0);
  }

  if (info->vpath)
    fprintf(fp, "#!gmake\n\n");

  time(&act_time);
  date = asctime(localtime(&act_time));
  if (pn = strstr(date, "\n"))
    for (*pn-- = 0; IS_BLANK_CHAR(*pn); pn--) *pn = 0;

  print_comment(fp, nil);
  sprintf(line, "Makefile %s for lib%s.a", info->makefile_name,info->name);
  print_comment(fp, line);
  sprintf(line, "automatically generated for ALBERT installation");
  print_comment(fp, line);
  sprintf(line, "by:       %s", getenv("USER"));
  print_comment(fp, line);
  sprintf(line, "on host:  %s", getenv("HOST")); 
  print_comment(fp, line);
  sprintf(line, "date:     %s", date);
  print_comment(fp, line);
  print_comment(fp, nil);
  NL(fp);

  fprintf(fp, ".PHONY:  lib\n");
  fprintf(fp, "default: lib\n\n");

  if (info->def_flags)  /* depend on DIM, DIM_OF_WORLD, etc.  */
  {
    print_comment(fp, nil);
    print_comment(fp, 
		  "Compile flags: DIM, DIM_OF_WORLD, NEIGH_IN_EL, EL_INDEX");
    print_comment(fp, nil);
    NL(fp);
    dflags = "$(DFLAGS)";
    fprintf(fp, "DFLAGS = -DDIM=$(DIM) ");
    fprintf(fp, "-DDIM_OF_WORLD=$(DIM_OF_WORLD)\\\n");
    fprintf(fp, "         -DNEIGH_IN_EL=$(NEIGH_IN_EL) ");
    fprintf(fp, "-DEL_INDEX=$(EL_INDEX)\n\n");
  }
  else
    dflags = nil;

  moflag = info->machine_oflag ? "$(MOFLAG)" : nil;

  print_comment(fp, nil);
  print_comment(fp, "CC, CFLAGS, FC, FFLAGS AR, ARFLAGS");
  print_comment(fp, nil);
  NL(fp);

  if (info->c_comp)
  {
    fprintf(fp, "CC = %s\n", info->c_comp);
    fprintf(fp, "CFLAGS = %s", ST(info->cflags));
    if (dflags)
      fprintf(fp, " %s", dflags);

    if (info->oflags)
      fprintf(fp, " $(OFLAGS)");
    if (moflag)
      fprintf(fp, " %s", moflag);
    if (info->vpath)
      fprintf(fp, " -I. -I%s", info->vpath);
  }
  NL(fp);
  if (info->f_comp)
  {
    fprintf(fp, "FC = %s\n", info->f_comp);
    fprintf(fp, "FFLAGS = %s", ST(info->fflags));
    if (info->oflags)
      fprintf(fp, " $(OFLAGS)");
    if (moflag)
      fprintf(fp, " %s", moflag);
  }
  NL(fp);
  NL(fp);

  if (info->vpath)
  {
    print_comment(fp, nil);
    print_comment(fp, "set virtual path");
    print_comment(fp, nil);
    NL(fp);

    fprintf(fp, "VPATH=./:%s\n", info->vpath);
    NL(fp);
  }

  fprintf(fp, "AR = ar\n");
/*   fprintf(fp, "ARFLAGS = rv\n\n"); */
  if (ranlib = get_parameter_value("RANLIB"))
    fprintf(fp, "RANLIB = %s\n\n", ranlib);

  print_comment(fp, nil);
  print_comment(fp, "Library path and name");
  print_comment(fp, nil);
  NL(fp);

  fprintf(fp, "LIB_PATH = %s\n", ST(info->lib_path));

  if (dflags)
    libext = "$(DIM)$(DIM_OF_WORLD)$(NEIGH_IN_EL)$(EL_INDEX)$(DEBUG)";
  else
    libext = "";

  lib_name = gen_lib_name(info->lib_name, "$(LIB_PATH)", libext, moflag);
  fprintf(fp, "LIB = %s\n", lib_name);
  NL(fp);

  print_comment(fp, nil);
  print_comment(fp, "Library members");
  print_comment(fp, nil);
  NL(fp);

  print_lib_files(info->ofiles, fp);

  NL(fp);

  print_comment(fp, nil);
  print_comment(fp, "Compile and archive command");
  print_comment(fp, nil);
  NL(fp);

  print_compile(info->ofiles, fp);

  NL(fp);

  print_comment(fp, nil);
  print_comment(fp, "Now, making the library");
  print_comment(fp, nil);
  NL(fp);

  fprintf(fp, "$(LIB): $(LIB_FILES)\n");
  if (ranlib)
    fprintf(fp, "\t$(RANLIB) $(LIB)\n");
  fprintf(fp, "\tchmod a+r $(LIB)\n\n");

  if (info->copy_header)
  {
    char        word[128];
    const char  *next = info->copy_header;
    while (next = get_word(next, word))
    {
      if (info->vpath)
      {
	fprintf(fp, "$(LIB_PATH)/%s: %s/%s\n", word, info->vpath, word);
	fprintf(fp, "\tcp %s/%s $(LIB_PATH)/%s\n", info->vpath, word, word);
      }
      else
      {
	fprintf(fp, "$(LIB_PATH)/%s: %s\n", word, word);
	fprintf(fp, "\tcp %s $(LIB_PATH)/%s\n", word, word);
      }
      fprintf(fp, "\tchmod a+r $(LIB_PATH)/%s\n", word);
    }
  }
  NL(fp);

  fprintf(fp, "lib: $(LIB)");
  if (info->copy_header)
  {
    char        word[128];
    const char  *next = info->copy_header;
    while (next = get_word(next, word))
      fprintf(fp, " $(LIB_PATH)/%s", word);
  }
  NL(fp);
  NL(fp);

  print_comment(fp, nil);
  NL(fp);

  fprintf(fp, "clean:\n\t-rm *.o *~\n");
  NL(fp);
  
  fclose(fp);

  return(1);
}

static void print_make_line(FILE *mkfi, const char *arch, 
			    const char *arguments, const char *oflags,
			    const char *moflags)
{
  char   value[128];

  if (moflags = get_word(moflags, value))
  {
    do
    {
      fprintf(mkfi, "\t$(MAKE) ");
      if (arch)
	fprintf(mkfi, "-f Makefile.%s", arch);
      if (oflags)
	fprintf(mkfi, " OFLAGS=\"%s\"", oflags);
      fprintf(mkfi, " MOFLAG=\"%s\"", value);
      if (arguments)
	fprintf(mkfi, "\\\n\t\t%s", arguments);
      if (moflags = get_word(moflags, value))
	fprintf(mkfi, ";\\\n");
      else
	fprintf(mkfi, "\n");
    } while(moflags);
  }
  else
  {
    fprintf(mkfi, "\t$(MAKE) ");
    if (arch)
      fprintf(mkfi, "-f Makefile.%s", arch);
    if (oflags)
      fprintf(mkfi, " OFLAGS=\"%s\"", oflags);
    if (arguments)
      fprintf(mkfi, "\\\n\t\t%s", arguments);
    fprintf(mkfi, "\n");
  }
}

static const char *init_defs(int dim, int dow, int nie, int ei, int debug)
{
  static char line[128];
  sprintf(line, "DIM=%d DIM_OF_WORLD=%d NEIGH_IN_EL=%d EL_INDEX=%d DEBUG=%d",
	  dim, dow, nie, ei, debug);
  return(line);
}

static void standard_makefile(const char *date)
{
  char line[128];
  const char *s, *a;
  FILE *mkfi;

  a = get_parameter_value("ALBERT_LIB_PATH");
  sprintf(line, "%s/Makefile", a);
  if (!(mkfi = fopen(line, "w")))
  {
    MSG("ERROR: can not open %s for writing\n", line);
    MSG("       %s is used by all ALBERT Makefiles!\n");
    MSG("       check read and write permissions\n");
    return;
  }

  print_comment(mkfi, nil);
  sprintf(line, "master makefile, included by application Makefiles");
  print_comment(mkfi, line);
  sprintf(line, "automatically generated during ALBERT installation");
  print_comment(mkfi, line);
  sprintf(line, "by:       %s", getenv("USER"));
  print_comment(mkfi, line);
  sprintf(line, "on host:  %s", getenv("HOST"));
  print_comment(mkfi, line);
  sprintf(line, "date:     %s", date);
  print_comment(mkfi, line);
  sprintf(line, "for architecture %s", get_parameter_value("ARCH"));
  print_comment(mkfi, nil);
  NL(mkfi);

  sprintf(line, "C and FORTRAN compilers and loader");
  print_comment(mkfi, nil);
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);

  fprintf(mkfi, "CC = %s\n", get_parameter_value("CC"));
  fprintf(mkfi, "FC = %s\n", get_parameter_value("FC"));
  fprintf(mkfi, "LD = %s\n", get_parameter_value("LD"));
  NL(mkfi);

  sprintf(line, "default settings");
  print_comment(mkfi, nil);
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);

  fprintf(mkfi, "MOFLAG = %s\n",
	  ST(get_parameter_value("DEFAULT_MACHINE_OFLAG")));
  if (s = get_parameter_value("CFLAGS"))
    fprintf(mkfi, "CFLAGS = %s\n", s);
  else
    fprintf(mkfi, "CFLAGS =\n");
  if (s = get_parameter_value("OFLAGS"))
    fprintf(mkfi, "OFLAG = %s\n", s);
  else
    fprintf(mkfi, "OFLAG =\n");
  s = get_parameter_value("DEBUG_LIB");
  if (strcmp(s, "0"))
    fprintf(mkfi, "DEBUG = 0\n");
  else
    fprintf(mkfi, "DEBUG = 1\n");
  NL(mkfi);

  sprintf(line, "used libraries");
  print_comment(mkfi, nil);
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);
  fprintf(mkfi, "BLAS_LIB = ");
  s = get_parameter_value("BLAS_LIB");
  if (*s)
  {
    s = get_parameter_value("BLAS_LIB_PATH");
    if (s && *s && strcmp(s, a))
      fprintf(mkfi, "-L%s ", s);
    fprintf(mkfi, "%s\n", get_parameter_value("BLAS_LIB"));
  }

  fprintf(mkfi, "GLTOOLS_LIB = ");
  if (atoi(get_parameter_value("CREATE_GLTOOLS_INTERFACE")))
  {
    s = get_parameter_value("GLTOOLS_LIB");
    if (s && *s)
    {
      s = get_parameter_value("GLTOOLS_LIB_PATH");
      if (s && *s && strcmp(s, a))
	fprintf(mkfi, "-L%s ", s);
      fprintf(mkfi, "%s$(MOFLAG)", get_parameter_value("GLTOOLS_LIB"));
    }
  }
  NL(mkfi);
  fprintf(mkfi, "PLOT_LIB = ");
  if ((s = get_parameter_value("PLOT_LIB")) && *s)
  {
    if ((s = get_parameter_value("PLOT_LIB_PATH")) && strcmp(s, a))
      fprintf(mkfi, "-L%s ", s);
    fprintf(mkfi, "%s$(MOFLAG) ", get_parameter_value("PLOT_LIB"));
  }
  NL(mkfi);
#if 0
  fprintf(mkfi, "GRAPE_LIB =\n");
#endif

  fprintf(mkfi, "GLX11_LIB = ");
  if (atoi(get_parameter_value("USE_OPENGL")))
  {
    if (s = get_parameter_value("OPENGL_LIB_PATH"))
    {
      if (s && *s && strcmp(s, a))
	fprintf(mkfi, "-L%s ", s);
    }
    if (s = get_parameter_value("OPENGL_LIB"))
    {
      if (*s)
	fprintf(mkfi, "%s ", s);
    }
  }
  if ((s = get_parameter_value("X11_LIB")) && *s)
    fprintf(mkfi, "%s\n", s);
  else
    NL(mkfi);

  NL(mkfi);
  sprintf(line, "define flags and compile commands");
  print_comment(mkfi, nil);
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);
  
  fprintf(mkfi, "DFLAGS = -DDIM=$(DIM)");
  fprintf(mkfi, " -DDIM_OF_WORLD=$(DIM_OF_WORLD)");
  fprintf(mkfi, " -DNEIGH_IN_EL=$(NEIGH_IN_EL)");
  fprintf(mkfi, " -DEL_INDEX=$(EL_INDEX)\n");
  fprintf(mkfi, ".c.o:\n");
  fprintf(mkfi, "\t$(CC) -c $(MOFLAG) $(OFLAG) $(CFLAGS) $(DFLAGS)");
  fprintf(mkfi, " -I$(ALBERT_LIB_PATH) $<\n");
  fprintf(mkfi, ".f.o:\n");
  fprintf(mkfi, "\t$(FC) -c $(MOFLAG) $(FFLAGS) $(OFLAG) $<\n");
  NL(mkfi);

  sprintf(line, "the ALBERT library and libraries used by ALBERT");
  print_comment(mkfi, nil);
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);

  fprintf(mkfi, "ALBERT_LIB = -lALBERT$(DIM)$(DIM_OF_WORLD)$(NEIGH_IN_EL)");
  fprintf(mkfi, "$(EL_INDEX)$(DEBUG)$(MOFLAG)\n");

  fprintf(mkfi, "LIBS = -L$(ALBERT_LIB_PATH) $(ALBERT_LIB) ");

  if ((s = get_parameter_value("SOLVER_LIB")) && *s)
  {
    if ((s = get_parameter_value("SOLVER_LIB_PATH")) && strcmp(s, a))
      fprintf(mkfi, "-L%s ", s);
    fprintf(mkfi, "%s$(MOFLAG) ", get_parameter_value("SOLVER_LIB"));
  }

#if 0
  fprintf(mkfi, "$(BLAS_LIB) $(GLTOOLS_LIB) $(PLOT_LIB) $(GRAPE_LIB) ");
  fprintf(mkfi, "$(GLX11_LIB) ");
#else
  fprintf(mkfi, "$(BLAS_LIB) $(GLTOOLS_LIB) $(PLOT_LIB) $(GLX11_LIB) ");
#endif
  if (strcmp(get_parameter_value("LD"), "f77"))
    fprintf(mkfi, "-lm");
  fprintf(mkfi, "\n");

  NL(mkfi);
  print_comment(mkfi, nil);
  print_comment(mkfi, "some often used targets ");
  print_comment(mkfi, nil);
  NL(mkfi);

  fprintf(mkfi, "default: $(DEFAULT)\n\n");

  fprintf(mkfi, ".PHONY: clean cleano new\n");
  fprintf(mkfi, "clean:\n\t-rm *.o *~\n\n");
  fprintf(mkfi, "cleano:\n\t-rm *.o\n\n");
  fprintf(mkfi, "new: cleano default\n");
  fprintf(mkfi, "\n");

  sprintf(line, "end of master makefile");
  print_comment(mkfi, nil);
  print_comment(mkfi, line);
  print_comment(mkfi, nil);

  fclose(mkfi);
  return;
}

static void application_makefile(int dim)
{
  char line[128];
  const char *a, *make, *s;
  FILE *mkfi;

  a = get_parameter_value("ALBERT_LIB_PATH");
  sprintf(line, "%s/Makefile.%dd", a, dim);
  if (!(mkfi = fopen(line, "w")))
  {
    MSG("ERROR: can not open %s for writing\n", line);
    MSG("       %s is a sample ALBERT Makefile for %dd!\n", dim);
    MSG("       check read and write permissions\n");
    return;
  }

  make = get_parameter_value("MAKE");
  if (strcmp("make", make))
    fprintf(mkfi, "#!%s\n", make);
  print_comment(mkfi, nil);
  sprintf(line, "Sample ALBERT Makefile for %dd", dim);
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);
  fprintf(mkfi, "DEFAULT = ellipt\n\n");

  fprintf(mkfi, "# delete line, if ALBERT_LIB_PATH is an environment variable\n");
  fprintf(mkfi, "ALBERT_LIB_PATH = %s\n\n", a);
  fprintf(mkfi, "include $(ALBERT_LIB_PATH)/Makefile\n\n");

  fprintf(mkfi, 
	  "# MOFLAG is machine dependent OFLAG -mips2-4 on SGIs, e.g.\n");
  fprintf(mkfi, "#MOFLAG = %s\n", 
	  ST(get_parameter_value("DEFAULT_MACHINE_OFLAG")));
  fprintf(mkfi, "# default compile flags\n");
  fprintf(mkfi, "#CFLAGS = %s\n", ST(get_parameter_value("CFLAGS")));
  fprintf(mkfi, "# default optimization flags\n");
  fprintf(mkfi, "#OFLAG = %s\n", ST(get_parameter_value("OFLAGS")));
  fprintf(mkfi, "# uncomment line for debugging\n");
  fprintf(mkfi, "#OFLAG = %s\n", ST(get_parameter_value("DEBUG_FLAGS")));
  s = get_parameter_value("DEBUG_LIB");
  if (strcmp(s, "1"))
  {
    fprintf(mkfi, "# uncomment line for using the debug library\n");
    fprintf(mkfi, "#DEBUG = 1\n");
  }
  NL(mkfi);
  print_comment(mkfi, nil);
  sprintf(line, "DFLAGS: DIM, DIM_OF_WORLD, NEIGH_IN_EL, EL_INDEX");
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);
  fprintf(mkfi, "DIM = %d\n", dim);
  fprintf(mkfi, "DIM_OF_WORLD = %d\n", dim);
  fprintf(mkfi, "EL_INDEX = 1\n");
  fprintf(mkfi, "NEIGH_IN_EL = 0\n\n");

  print_comment(mkfi, nil);
  print_comment(mkfi, "set virtal path");
  print_comment(mkfi, nil);
  fprintf(mkfi, "VPATH=./:../Common");

  print_comment(mkfi, nil);
  sprintf(line, "and now the user's files ");
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);
  fprintf(mkfi, "ELLIPT_OFILES = ellipt.o\n\n");
  fprintf(mkfi, "ellipt: $(ELLIPT_OFILES)\n");
  fprintf(mkfi, "\t$(LD) $(MOFLAG) -o $@ $(ELLIPT_OFILES) $(LIBS)\n");

  return;
}

main(int argc, const char *argv[])
{
  int            n_dim,      dim[3];
  int            n_dow,      dow[5];
  int            n_neigh,    neigh[2];
  int            n_el_index, el_index[2];
  int            n_debug,    debug[2];
  int            i, j, k, l;
  const char     *moflags, *defs, *arch;
  const char     *plot_lib = nil, *solver_lib = nil;
  const char     *blas_lib = nil, *albert_lib = nil;
  const char     *oflags, *dflags, *s;
  char           line[128], *pn, *date;
  time_t         act_time;
  MAKEFILE_INFO  plot_info[1] = {{nil}};
  MAKEFILE_INFO  blas_info[1] = {{nil}};
  MAKEFILE_INFO  solver_info[1] = {{nil}};
  MAKEFILE_INFO  albert_info[4] = {{nil}};
  FILE           *mkfi;
  const char     *makefile_name = "Makefile";
  const char     *parfile = argc > 1 ? argv[1] : nil;

  funcName = argv[0];
  if (argc < 2)
  {
    MSG("usage %s datfile\n", argv[0]);
    MSG("writing a default file: default.dat\n");
    MSG("modify this file to your needs\n");
    if (mkfi = fopen("default.dat", "w"))
	print_parameters(mkfi, 1, 1);
    exit(0);
  }

  if (init_parameters(parfile)) exit(1);  /*  errors during par. reading */
  sprintf(line, "%s.used", parfile);

  if (mkfi = fopen(line, "w"))
  {
    print_parameters(mkfi, 0, 0);
    fclose(mkfi);
  }

  if (!(mkfi = fopen(makefile_name, "w")))
  {
    MSG("ERROR: can not open %s for writing\n", makefile_name);
    MSG("       check read and write permissions\n");
    exit(1);
  }

  time(&act_time);
  date = asctime(localtime(&act_time));
  if (pn = strstr(date, "\n"))
    for (*pn-- = 0; IS_BLANK_CHAR(*pn); pn--) *pn = 0;

  print_comment(mkfi, nil);
  sprintf(line, "Makefile %s for creating ALBERT libs", makefile_name);
  print_comment(mkfi, line);
  sprintf(line, "automatically generated for ALBERT installation");
  print_comment(mkfi, line);
  sprintf(line, "by:       %s", getenv("USER"));
  print_comment(mkfi, line);
  sprintf(line, "on host:  %s", getenv("HOST"));
  print_comment(mkfi, line);
  sprintf(line, "date:     %s", date);
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);
  NL(mkfi);

  arch = get_parameter_value("ARCH");
  sprintf(line, "default architecture %s", arch);
  print_comment(mkfi, nil);
  print_comment(mkfi, line);
  print_comment(mkfi, nil);
  NL(mkfi);
  fprintf(mkfi, "ARCH = %s\n", arch);
  NL(mkfi);
  arch = "$(ARCH)";

  fprintf(mkfi, "default: ");
  if (atoi(get_parameter_value("CREATE_PLOT_LIB")))
    fprintf(mkfi, "%s ", plot_lib = "PLOTlib");
  if (atoi(get_parameter_value("CREATE_SOLVER_LIB")))
    fprintf(mkfi, "%s ", solver_lib = "SOLVERlib");
  if (atoi(get_parameter_value("CREATE_BLAS_LIB")))
    fprintf(mkfi, "%s ", blas_lib = "BLASlib");
  if (atoi(get_parameter_value("CREATE_ALBERT_LIB")))
  {
    albert_lib = "ALBERTlib";
    n_dim = fill_vector(get_parameter_value("DIM"), dim, 3);
    fprintf(mkfi, "%s", albert_lib);
  }
  fprintf(mkfi, "\n");
  if (albert_lib)
  {
    fprintf(mkfi, "%s: ", albert_lib);
    for (i = 0; i < n_dim; i++)
      fprintf(mkfi, "%s%dd ", albert_lib, dim[i]);
  }
  fprintf(mkfi, "\n");
  fprintf(mkfi, ".PHONY: ");
  if (plot_lib)
    fprintf(mkfi, "%s ", plot_lib);
  if (solver_lib)
    fprintf(mkfi, "%s ", solver_lib);
  if (blas_lib)
    fprintf(mkfi, "%s ", blas_lib);
  if (albert_lib)
  {
    fprintf(mkfi, "%s ", albert_lib);
    for (i = 0; i < n_dim; i++)
      fprintf(mkfi, "%s%dd ", albert_lib, dim[i]);
  }
  fprintf(mkfi, "tar ");
  fprintf(mkfi, "clean\n\n");


  print_comment(mkfi, nil);
  print_comment(mkfi, "make makefile program mkfi");
  print_comment(mkfi, nil);
  NL(mkfi);

  fprintf(mkfi, "CC = %s\n", get_parameter_value("CC"));
  fprintf(mkfi, "mkfi: mkfi.c\n");
  fprintf(mkfi, "\t$(CC) -o mkfi mkfi.c\n\n");

  print_comment(mkfi, nil);
  print_comment(mkfi, "make commands");
  print_comment(mkfi, nil);
  NL(mkfi);

  fprintf(mkfi, "MAKE = %s\n\n", get_parameter_value("MAKE"));

/****************************************************************************/
/* make makefile for plot lib                                               */
/****************************************************************************/

  if (plot_lib)
  {
    init_plot_files(plot_info);
    oflags = get_parameter_value("OFLAGS");

    print_comment(mkfi, nil);
    print_comment(mkfi, "Commands for making the plot lib");
    print_comment(mkfi, nil);
    NL(mkfi);

    make_makefile(plot_info);
    fprintf(mkfi, "%s:\n", plot_lib);
    fprintf(mkfi, "\t@echo Making %s\n", plot_lib);
    fprintf(mkfi, "\t@cd %s;\\\n", plot_info->source_path);

    moflags = plot_info->machine_oflag;
    print_make_line(mkfi, arch, nil, oflags, moflags);
    NL(mkfi);
  }

/****************************************************************************/
/* make makefile for solver lib                                             */
/****************************************************************************/

  if (solver_lib)
  {
    init_csolver_files(solver_info);
    oflags = get_parameter_value("OFLAGS");

    print_comment(mkfi, nil);
    print_comment(mkfi, "Commands for making the C-solver lib");
    print_comment(mkfi, nil);
    NL(mkfi);

    make_makefile(solver_info);
    fprintf(mkfi, "%s:\n", solver_lib);
    fprintf(mkfi, "\t@echo Making %s\n", solver_lib);
    fprintf(mkfi, "\t@cd %s;\\\n", solver_info->source_path);

    moflags = solver_info->machine_oflag;
    print_make_line(mkfi, arch, nil, oflags, moflags);
    NL(mkfi);
  }

/****************************************************************************/
/* make makefile for blas lib                                               */
/****************************************************************************/

  if (blas_lib)
  {
    init_blas_files(blas_info);
    oflags = get_parameter_value("OFLAGS");

    print_comment(mkfi, nil);
    print_comment(mkfi, "Commands for making the blas lib");
    print_comment(mkfi, nil);
    NL(mkfi);

    make_makefile(blas_info);
    fprintf(mkfi, "%s:\n", blas_lib);
    fprintf(mkfi, "\t@echo Making %s\n", blas_lib);
    fprintf(mkfi, "\t@cd %s;\\\n", blas_info->source_path);

    moflags = blas_info->machine_oflag;
    print_make_line(mkfi, arch, nil, oflags, moflags);
    NL(mkfi);
  }


/****************************************************************************/
/* make makefiles for ALBERT libs                                           */
/****************************************************************************/

  if (albert_lib)
  {
    int  d;

    init_albert_files(albert_info);
    n_neigh    = fill_vector(get_parameter_value("NEIGH_IN_EL"), neigh, 2);
    n_el_index = fill_vector(get_parameter_value("EL_INDEX"), el_index, 2);
    n_debug    = fill_vector(get_parameter_value("DEBUG_LIB"), debug, 2);
    s = get_parameter_value("ALBERT_SOURCE_PATH");

    for (d = 0; d < n_dim; d++)
    {
      print_comment(mkfi, nil);
      sprintf(line, "Commands for making the ALBERT lib for %1d", dim[d]);
      print_comment(mkfi, line);
      print_comment(mkfi, nil);
      NL(mkfi);

      make_makefile(&albert_info[dim[d]]);
      fprintf(mkfi, "%s%1dd:\n", albert_lib, dim[d]);
      fprintf(mkfi, "\t@echo Making %s%1dd\n", albert_lib, dim[d]);
      
      sprintf(line, "DIM_OF_WORLD_%1dD", dim[d]);
      n_dow = fill_vector(get_parameter_value(line), dow, 5);
      moflags = albert_info[dim[d]].machine_oflag;
      oflags = get_parameter_value("OFLAGS");
      dflags = get_parameter_value("DEBUG_FLAGS");

      for (i = 0; i < n_dow; i++)   /*  loop dow  */
      {
	for (j = 0; j < n_neigh; j++)  /*  loop neigh_in_el  */
	{
	  for (k = 0; k < n_el_index; k++)  /*  loop el_index  */
	  {
	    for (l = 0; l < n_debug; l++)   /*  loop debug  */
	    {
	      defs = init_defs(dim[d], dow[i], neigh[j], el_index[k], 
			       debug[l]);
	      fprintf(mkfi, "\techo lib for %s;\\\n", defs);
	      fprintf(mkfi, "\tcd %s;\\\n", albert_info[dim[d]].source_path);
	      if (debug[l])
	      {
		print_make_line(mkfi, arch, defs, dflags, moflags);
	      }
	      else
		print_make_line(mkfi, arch, defs, oflags, moflags);
	    }
	  }
	}
      }
      application_makefile(dim[d]);
    }
    standard_makefile(date);

    if (n_debug)
    {
      fprintf(mkfi, "\t-@cd %s; \\\n", s);
      fprintf(mkfi, "chmod a+r */*.c */*.h\n");
    }
    NL(mkfi);
  }
  NL(mkfi);

  print_comment(mkfi, nil);
  print_comment(mkfi, "command for tar files");
  print_comment(mkfi, nil);

  fprintf(mkfi, "tar: ");
  s = get_parameter_value("ALBERT_SOURCE_PATH");
  if (s && *s) fprintf(mkfi, "albert.tar ");
  s = get_parameter_value("SOLVER_SOURCE_PATH");
  if (s && *s) fprintf(mkfi, "solver.tar ");
  s = get_parameter_value("PLOT_SOURCE_PATH");
  if (s && *s) fprintf(mkfi, "plot.tar ");
  NL(mkfi);  NL(mkfi);

  s = get_parameter_value("ALBERT_SOURCE_PATH");
  if (s && *s)
  {
    sprintf(line, "%s/albert.tar", getenv("PWD"));
    fprintf(mkfi, "albert.tar:\n");
    fprintf(mkfi, "\t@echo Making albert.tar\n");
    fprintf(mkfi, "\t@cd %s;\\\n", s);
    fprintf(mkfi, "\ttar cvf %s ", line);
    fprintf(mkfi, "*d/*.c *d/*.h Common/*.c Common/*.h\n");
  }

  s = get_parameter_value("SOLVER_SOURCE_PATH");
  if (s && *s)
  {
    sprintf(line, "%s/solver.tar", getenv("PWD"));
    fprintf(mkfi, "solver.tar:\n");
    fprintf(mkfi, "\t@echo Making solver.tar\n");
    fprintf(mkfi, "\t@cd %s;\\\n", s);
    fprintf(mkfi, "\ttar cvf %s *.c *.h *.f\n", line);
  }

  s = get_parameter_value("PLOT_SOURCE_PATH");
  if (s && *s)
  {
    sprintf(line, "%s/plot.tar", getenv("PWD"));
    fprintf(mkfi, "plot.tar:\n");
    fprintf(mkfi, "\t@echo Making plot.tar\n");
    fprintf(mkfi, "\t@cd %s;\\\n", s);
    fprintf(mkfi, "\ttar cvf %s *.c *.h\n", line);
  }

  print_comment(mkfi, nil);
  print_comment(mkfi, "command for cleaning up");
  print_comment(mkfi, nil);
  
  fprintf(mkfi, "clean:\n");
  fprintf(mkfi, "\t-/bin/rm mkfi\n");
  fprintf(mkfi, "\tfind . \\( -name \\*~ -o -name \\*.o \\)");
  fprintf(mkfi, " -exec /bin/rm -f {} \\;\n");

  print_comment(mkfi, nil);
  print_comment(mkfi, "command for making tgz file");
  print_comment(mkfi, nil);

  fprintf(mkfi, ".PHONY: tgz\n");
  fprintf(mkfi, "DIST_DATA = AUTHORS COPYING INSTALL NEWS README");
  fprintf(mkfi, " mkfi.c *.dat\\\n");
  fprintf(mkfi, "            ");
  fprintf(mkfi, " LIB_SOURCE/*/*.[ch] PLOT_ANSI/*.[ch] SOLVER/*.[chf]");
  fprintf(mkfi, " demo.tar\n\n");

  fprintf(mkfi, "tgz:\n");
  fprintf(mkfi, "\tcd Demo; make tar; mv demo.tar ..\n");
  fprintf(mkfi, "\ttar cvf - $(DIST_DATA) | gzip -9v > albert-1.1.tgz\n");
  fprintf(mkfi, "\t@ls -l albert-1.1.tgz\n");
  fclose(mkfi);
}
