/****************************************************************************/
/* ALBERT_UTIL:  tools for messages, memory allocation, parameters, etc.    */
/*                                                                          */
/* file:     divers.c                                                       */
/*                                                                          */
/* description:  generating filenames and handling of general lists         */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-2000)                       */
/*                                                                          */
/****************************************************************************/

#include "albert_util.h"

const char *generate_filename(const char *path, const char *fn, int ntime)
{
  static char  name[256];
  char         *cp;
  
  if (path == nil || path[0] == '\0')
  {
    sprintf(name, "./%s", fn);
  }
  else
  {
    const char *cp = path;
    while (*cp)
      cp++;
    cp--;
    if (*cp == '/')
      sprintf(name, "%s%s", path, fn);
    else
      sprintf(name, "%s/%s", path, fn);
  }
  cp = name;
  while (*cp)
    cp++;
  sprintf(cp, "%06d", ntime);

  return((const char *) name);
}

/****************************************************************************/
/*  handling of general lists                                               */
/****************************************************************************/

static VOID_LIST_ELEMENT *first_element = nil;

VOID_LIST_ELEMENT *get_void_list_element()
{
  VOID_LIST_ELEMENT  *list;
  const int          n_elements = 20;
  int                i;
  
  if (!first_element)
  {
    first_element = MEM_ALLOC(n_elements, VOID_LIST_ELEMENT);
    for (i = 0; i < n_elements-1; i++)
    {
      first_element[i].data = nil;
      first_element[i].next = first_element+(i+1);
    }
    first_element[n_elements-1].data = nil;
    first_element[n_elements-1].next = nil;
  }
  list = first_element;
  first_element = list->next;

  list->next = nil;
  return(list);
}

void free_void_list_element(VOID_LIST_ELEMENT *list)
{
  list->data = nil;
  list->next = first_element;
  first_element = list;
  return;
}
