#include <albert.h>

/****************************************************************************/
/*  graphics:                                                               */
/*  display of mesh elements and mesh element marker                        */
/****************************************************************************/
#define USE_GTOOLS 1

#if USE_GTOOLS
static void graphics(MESH *mesh, int mark)
{
  FUNCNAME("graphics");
  static GLTOOLS_WINDOW   win = nil;

  if (!mesh) 
    return;

  if (!win)
   win = open_gltools_window("ALBERT", "500x500+0+0", nil, mesh, 1);

  if (win) gltools_mesh(win, mesh, mark);

  return;
}
#else
#if DIM == 2
static void graphics(MESH *mesh, int mark)
{
  FUNCNAME("graphics");
  static GRAPH_WINDOW  win = nil;

  if (!mesh) 
    return;

  if (!win)
   win = graph_open_window("ALBERT", "300x300+0+0", nil, mesh);

  if (win)
  {
    graph_clear_window(win, nil);
    if (mark)
      graph_mesh(win, mesh, nil, GRAPH_MESH_ELEMENT_MARK);
    else
      graph_mesh(win, mesh, nil, 0);
  }
  graph_mesh(win, mesh, nil, GRAPH_MESH_BOUNDARY);
  WAIT;
  return;
}
#endif
#endif
#include <time.h>
static int random_ref_mark;
static void random_refine_fct(const EL_INFO *el_info)
{
  FUNCNAME("random_refine_fct");

  if (drand48() <= 0.5)
    el_info->el->mark = random_ref_mark;
}

/****************************************************************************/
/*  random_refine:							    */
/*  sets randomly refinement marks on the elements and refines the mesh     */
/*  drand48 is initialized by the global variable time_is_cash_time_is_money*/
/*  defined in femtest.c						    */
/****************************************************************************/

void random_refine_mark(MESH *mesh, int mark)
{
  FUNCNAME("random_refine");
  static int      first = 1;
  time_t          act_time;

  random_ref_mark = mark;

  if (first--)
  {
    time(&act_time);
    srand48(act_time);
  }

  mesh_traverse(mesh, -1, CALL_LEAF_EL, random_refine_fct);
}
static int random_coars_mark;

static void random_coarsen_fct(const EL_INFO *el_info)
{
  FUNCNAME("random_coarsen_fct");

  if (drand48() <= 0.5)
    el_info->el->mark = -random_coars_mark;
}

/****************************************************************************/
/*  random_coarsen:							    */
/*  sets randomly coarsening marks on the elements and coarsens the mesh    */
/****************************************************************************/

void random_coarsen_mark(MESH *mesh, int mark)
{
  FUNCNAME("random_coarsen_mark");
  static int      first = 1;

  random_coars_mark = mark;
  mesh_traverse(mesh, -1, CALL_LEAF_EL, random_coarsen_fct);
}

static void test(MESH *mesh)
{
  random_refine_mark(mesh, 5);
  random_coarsen_mark(mesh, 5);
  graphics(mesh, 1);
  refine(mesh);
  graphics(mesh, 1);
  coarsen(mesh);
  graphics(mesh, 1);
}

/****************************************************************************/
/*  an example for global refinement                                        */
/****************************************************************************/

static int refine_global_mark;

static void refine_global_fct(const EL_INFO *el_info)
{
  FUNCNAME("refine_global_fct");

  el_info->el->mark = refine_global_mark;
}

static void refine_global(MESH *mesh, int mark)
{
  FUNCNAME("refine_global");

  refine_global_mark = mark;

  mesh_traverse(mesh, -1, CALL_LEAF_EL, refine_global_fct);
  graphics(mesh, 1);

  refine(mesh);
  graphics(mesh, 1);
  return;
}

int main(int argc, char **argv)
{
  FUNCNAME("main");
  MESH     *mesh;
  int      n_refine;

/****************************************************************************/
/*  first of all, init parameters of the init file                          */
/****************************************************************************/

  init_parameters(1, "INIT/init.dat");
  
/****************************************************************************/
/*  get a mesh, and                                                         */
/*  read the macro triangulation from file                                  */
/****************************************************************************/

  mesh = GET_MESH("my first mesh", nil, nil);
  mesh->preserve_coarse_dofs = 0;
  read_macro(mesh, "Macro/macro_big.stand", nil);

  n_refine = 2;
  GET_PARAMETER(2, "global refinements", "%d", &n_refine);

  refine_global(mesh, n_refine*DIM);
  test(mesh);

  WAIT;
  return(0);
}
