#include "albert.h"

/****************************************************************************/
/* graphical output of mesh, discrete solution, and estimate                */
/****************************************************************************/

/*
#define HAVE_LIBGLTOOLS false
*/

#if HAVE_LIBGLTOOLS
void graphics(MESH *mesh, DOF_REAL_VEC *u_h, REAL (*get_est)(EL *el))
{
  FUNCNAME("graphics");
  static int first = true;
  static GLTOOLS_WINDOW   win_est = nil, win_val = nil;

  if (first)
  {
    int size[2] = {0, 0};
    char   geom[128] = "500x500+0+0";
    GET_PARAMETER(1, "graphic windows", "%d %d", size, size+1);

    if (size[0] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[0], size[0], 0, 0); 
      win_val = open_gltools_window("ALBERT values", geom, nil, mesh, true);
    }

    if (size[1] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[1], size[1], size[0], 0); 
      win_est = open_gltools_window("ALBERT estimate", geom, nil, mesh, true);
    }
    first = false;
  }

  if (u_h && win_val)
    gltools_drv(win_val, u_h, 1.0, 0.0);
  if (get_est && win_est)
    gltools_est(win_est, mesh, get_est, 0.0, -1.0);
  if (mesh && win_val && !u_h  && !get_est)
    gltools_mesh(win_val, mesh, 1);  /*  graph mesh with element marker   */

  if (!win_val && !win_est)  WAIT;

  return;
}
#else
#if DIM == 1
void graphics(MESH *mesh, DOF_REAL_VEC *u_h, REAL (*get_est)(EL *el))
{
  FUNCNAME("graphics");
  static int first = true;
  static GRAPH_WINDOW win_est = nil, win_val = nil, win_mesh = nil;

  if (first)
  {
    int size[3] = {0, 0, 0};
    char   geom[128] = "500x500+0+0";
    GET_PARAMETER(1, "graphic windows", "%d %d %d", size, size+1, size+2);

    if (size[0] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[0], size[0], 0, 0); 
      win_val = graph_open_window("ALBERT values", "300x300+305+0",nil,mesh);
    }
    if (size[1] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[0], size[0], size[0], 0); 
      win_est = graph_open_window("ALBERT estimate", "300x300+610+0",nil,mesh);
    }
    if (size[2] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[2], size[2], size[1], 0); 
      win_mesh = graph_open_window("ALBERT mesh", "300x300+0+0",nil,mesh);
    }
    first = false;
  }

  if (mesh && win_mesh)
  {
    graph_clear_window(win_mesh, rgb_white);
    graph_mesh(win_mesh, mesh, rgb_black, GRAPH_MESH_VERTEX_DOF);
  }
  if (u_h && win_val)
  {
    graph_clear_window(win_val, rgb_white);
    graph_mesh(win_val, u_h->fe_space->mesh, rgb_black, 0);
    graph_drv(win_val, u_h, rgb_black, 0.0, 0.0);
  }

  if (get_est && win_est)
  {
    graph_clear_window(win_est, rgb_black);
/*     graph_el_est(win_est, mesh, get_est); */
  }
  WAIT;

  return;
}
#endif
#if DIM == 2
void graphics(MESH *mesh, DOF_REAL_VEC *u_h, REAL (*get_est)(EL *el))
{
  FUNCNAME("graphics");
  static int first = true;
  static GRAPH_WINDOW win_est = nil, win_val = nil, win_mesh = nil;

  if (first)
  {
    int size[3] = {0, 0, 0};
    char   geom[128] = "500x500+0+0";
    GET_PARAMETER(1, "graphic windows", "%d %d %d", size, size+1, size+2);

    if (size[0] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[0], size[0], 0, 0); 
      win_val = graph_open_window("ALBERT values", "300x300+305+0",nil,mesh);
    }
    if (size[1] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[0], size[0], size[0], 0); 
      win_est = graph_open_window("ALBERT estimate", "300x300+610+0",nil,mesh);
    }
    if (size[2] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[2], size[2], size[1], 0); 
      win_mesh = graph_open_window("ALBERT mesh", "300x300+0+0",nil,mesh);
    }
    first = false;
  }

  if (mesh && win_mesh)
  {
    MSG("drawing mesh\n");
    graph_clear_window(win_mesh, rgb_white);
    graph_mesh(win_mesh, mesh, rgb_black, GRAPH_MESH_VERTEX_DOF);
  }
  if (u_h && win_val)
  {
    graph_clear_window(win_val, rgb_black);
    graph_values(win_val, u_h, 0);
  }
  if (get_est && win_est)
  {
    graph_clear_window(win_est, rgb_black);
    graph_el_est(win_est, mesh, get_est);
  }
  WAIT;

  return;
}
#endif
#if DIM == 3
void graphics(MESH *mesh, DOF_REAL_VEC *u_h, REAL (*get_est)(EL *el))
{
  FUNCNAME("graphics");
  MSG("only in 1d and 2d available, please use gltools graphics\n");
}
#endif
#endif
