#include "nonlin.h"

struct op_info
{
  REAL_D  Lambda[DIM+1];    /*  the gradient of the barycentric coordinates */
  REAL    det;              /*  det |D F_S|                                 */

  REAL    k_heat, sigma;

  const QUAD_FAST *quad_fast;   /*  quad_fast for the 0 order term!      */
  const REAL      *uk_qp;
};

static const REAL (*LALt(const EL_INFO *el_info, const QUAD *quad, 
			 int iq, void *ud))[DIM+1]
{
  struct op_info *info = ud;
  REAL           fac = info->k_heat*info->det;
  int            i, j, k;
  static REAL    LALt[DIM+1][DIM+1];

  for (i = 0; i <= DIM; i++)
    for (j = i; j <= DIM; j++)
    {
      for (LALt[i][j] = k = 0; k < DIM_OF_WORLD; k++)
	LALt[i][j] += info->Lambda[i][k]*info->Lambda[j][k];
      LALt[i][j] *= fac;
      LALt[j][i] = LALt[i][j];
    }
  return((const REAL (*)[DIM+1]) LALt);
}

static REAL c(const EL_INFO *el_info, const QUAD *quad, int iq, void *ud)
{
  struct op_info *info = ud;
  REAL  uk3;

  TEST_EXIT(info->quad_fast->quad == quad)("quads differ\n");

  uk3 = info->uk_qp[iq]*info->uk_qp[iq]*info->uk_qp[iq];

  return(info->sigma*info->det*uk3);
}

static void assemble(DOF_REAL_VEC *uk, DOF_MATRIX *matrix, DOF_REAL_VEC *fh,
		     REAL k_heat, REAL sigma, REAL (*f)(const REAL_D))
{
  FUNCNAME("assemble");
  static struct op_info *op_info = nil;
  static const REAL     **(*fill_a)(const EL_INFO *, void *) = nil;
  static void           *a_info = nil;
  static const REAL     **(*fill_c)(const EL_INFO *, void *) = nil;
  static void           *c_info = nil;
  static REAL           **el_mat;
  static int            el_mat_size = 0;

  static const DOF_ADMIN *admin = nil;
  static int             n_bas_fcts;
  static const REAL   *(*get_uk_loc)(const EL *, const DOF_REAL_VEC *, REAL *);
  static const DOF    *(*get_dof)(const EL *, const DOF_ADMIN *, DOF *);
  static const S_CHAR *(*get_bound)(const EL_INFO *, S_CHAR *);

  TRAVERSE_STACK *stack = get_traverse_stack();
  EL_INFO        *el_info;
  FLAGS          fill_flag;
  const REAL     **a_mat, **c_mat;
  REAL           *f_vec;
  const QUAD     *quad;
  int            i, j;

  quad = get_quadrature(DIM, 2*uk->fe_space->bas_fcts->degree - 2);

/****************************************************************************/
/*  init functions for matrix assembling                                    */
/****************************************************************************/

  if (admin != uk->fe_space->admin)
  {
    OPERATOR_INFO          o_info2 = {nil}, o_info0 = {nil};
    const EL_MATRIX_INFO   *matrix_info;
    const BAS_FCTS         *bas_fcts = uk->fe_space->bas_fcts;

    admin      = uk->fe_space->admin;
    if (el_mat_size < bas_fcts->n_bas_fcts)
    {
      if (el_mat_size)  MAT_FREE(el_mat, el_mat_size, el_mat_size, REAL);
      el_mat_size = bas_fcts->n_bas_fcts;
      el_mat = MAT_ALLOC(el_mat_size, el_mat_size, REAL);
    }
    n_bas_fcts = bas_fcts->n_bas_fcts;
    get_dof    = bas_fcts->get_dof_indices;
    get_bound  = bas_fcts->get_bound;
    get_uk_loc = bas_fcts->get_real_vec;;

    if (!op_info)
      op_info = MEM_ALLOC(1, struct op_info);

    o_info2.row_fe_space = o_info2.col_fe_space = uk->fe_space;

    o_info2.quad[2]        = quad;
    o_info2.LALt           = LALt;
    o_info2.LALt_pw_const  = true;
    o_info2.LALt_symmetric = true;
    o_info2.user_data      = op_info;

    matrix_info = fill_matrix_info(&o_info2, nil);
    fill_a = matrix_info->el_matrix_fct;
    a_info = matrix_info->fill_info;

    o_info0.row_fe_space = o_info0.col_fe_space = uk->fe_space;

    o_info0.quad[0]        = quad;
    o_info0.c              = c;
    o_info0.c_pw_const     = false;
    o_info0.user_data      = op_info;

    op_info->quad_fast = get_quad_fast(bas_fcts, quad, INIT_PHI);

    matrix_info = fill_matrix_info(&o_info0, nil);
    fill_c = matrix_info->el_matrix_fct;
    c_info = matrix_info->fill_info;
  }

/****************************************************************************/
/*  and now assemble the matrix and right hand side                         */
/****************************************************************************/
  op_info->k_heat = k_heat;
  op_info->sigma  = sigma;

  clear_dof_matrix(matrix);
  dof_set(0.0, fh);
  L2scp_fct_bas(f, quad, fh);
  f_vec = fh->vec;

  fill_flag = CALL_LEAF_EL|FILL_COORDS|FILL_BOUND;
  el_info = traverse_first(stack, uk->fe_space->mesh, -1, fill_flag);
  while (el_info)
  {
    const REAL   *uk_loc = (*get_uk_loc)(el_info->el, uk, nil);
    const DOF    *dof    = (*get_dof)(el_info->el, admin, nil);
    const S_CHAR *bound  = (*get_bound)(el_info, nil);

/****************************************************************************/
/* initialization of values used by LALt and c                              */
/****************************************************************************/
    op_info->det   = el_grd_lambda(el_info, op_info->Lambda);
    op_info->uk_qp = uh_at_qp(op_info->quad_fast, uk_loc, nil);

    a_mat = fill_a(el_info, a_info);
    c_mat = fill_c(el_info, c_info);

/****************************************************************************/
/*  add a(psi_i,psi_j) + m(4*u^3*psi_i,psi_j) to system matrix              */
/****************************************************************************/

    for (i = 0; i < n_bas_fcts; i++)
      for (j = 0; j < n_bas_fcts; j++)
	el_mat[i][j] = a_mat[i][j] + 4.0*c_mat[i][j];

    add_element_matrix(matrix, 1.0, n_bas_fcts, n_bas_fcts, dof, dof, 
		       (const REAL **) el_mat, bound);

/****************************************************************************/
/*  f += - a(uk,psi_i) - m(uk^4,psi_i)                                      */
/****************************************************************************/

    for (i = 0; i < n_bas_fcts; i++)
    {
      if (bound[i] < DIRICHLET)
      {
	REAL val = 0.0;
	for (j = 0; j < n_bas_fcts; j++)
	  val += (a_mat[i][j] + c_mat[i][j])*uk_loc[j];
	f_vec[dof[i]] -= val;
      }
      else
	f_vec[dof[i]] = 0.0;
    }

    el_info = traverse_next(stack, el_info);
  }

  free_traverse_stack(stack);

  return;
}

int newton(DOF_REAL_VEC *u0, REAL k_heat, REAL sigma, REAL (*f)(const REAL_D),
	   REAL tolerance, int max_iter, int info)
{
  FUNCNAME("newton");
  static DOF_REAL_VEC *fh = nil;
  static DOF_REAL_VEC *v = nil;
  static DOF_MATRIX   *A = nil;
  static REAL         ltol;
  static int          lmiter = 500, licon = 1, linfo = 4, lrestart = 10;
  int                 iter;
  REAL                resid = 0.0;
  DOF_REAL_VEC        *uk = u0;
  OEM_SOLVER          solver;

  if (!fh)
  {
    fh = get_dof_real_vec("fh", u0->fe_space);
    v  = get_dof_real_vec("v", u0->fe_space);
    A  = get_dof_matrix("A", u0->fe_space);

    ltol = 1.e-2*tolerance;
    GET_PARAMETER(1, "linear solver tolerance", "%f", &ltol);
    GET_PARAMETER(1, "linear solver max iteration", "%d", &lmiter);
    GET_PARAMETER(1, "linear solver info", "%d", &linfo);
    GET_PARAMETER(1, "linear solver precon", "%d", &licon);
    GET_PARAMETER(1, "linear solver restart", "%d", &lrestart);
  }

  for (iter = 0; iter < max_iter; iter++)
  {
    assemble(uk, A, fh, k_heat, sigma, f);

    dof_set(0.0, v);
/*--- if u_k->vec[i] >= 0 for all i => matrix is positive definite (p=1)  --*/
    solver = dof_min(uk) >= 0 ? CG : GMRes;
    oem_solve_s(A, fh, nil, v, solver, ltol, licon, lrestart, lmiter, linfo);

    dof_axpy(1.0, v, uk);

    if ((resid = H1_norm_uh(nil, v)) <= tolerance)
    {
      INFO(info,4)("finished successfully after %d iterations\n", iter+1);
      INFO(info,4)("resid: %le\n", resid);
      return(iter+1);
    }

    INFO(info,6)("iteration %2d: resid: %le\n", iter+1, resid);
  }

  INFO(info,2)("tolerance %lf not reached after %d iterations\n", 
	       tolerance, iter+1);
  INFO(info,2)("last residual = %lf\n", resid);
  return(max_iter+1);
}
