#include <albert.h>

/****************************************************************************/
/*  structure for collecting data of the problem:                           */
/*    k:     consant heat conductivity                                      */
/*    sigma: Stefan--Bolzmann constant                                      */
/*    g:     pointer to a function for evaluating boundary values           */
/*    f:     pointer to a function for evaluating the right-hand side       */
/*           (heat source)                                                  */
/*    u0:    if not nil, pointer to a function for evaluating an initial    */
/*           guess for the discrete solution on the macro triangulation     */
/*    u:     if not nil, pointer to a function for evaluating the true      */
/*           solution (only for test purpose)                               */
/*    grd_u: if not nil, pointer to a function for evaluating the gradient  */
/*           of the true solution (only for test purpose)                   */
/****************************************************************************/

typedef struct prob_data PROB_DATA;
struct prob_data
{
  REAL          k, sigma;

  REAL          (*g)(const REAL_D x);
  REAL          (*f)(const REAL_D x);

  REAL          (*u0)(const REAL_D x);

  REAL          (*u)(const REAL_D x);
  const REAL    *(*grd_u)(const REAL_D x);
};

/*--- file nlprob.c --------------------------------------------------------*/
const PROB_DATA *init_problem(MESH *mesh);

/*--- file nlsolve.c -------------------------------------------------------*/
int nlsolve(DOF_REAL_VEC *, REAL, REAL, REAL (*)(const REAL_D));

/*--- file graphics.c ------------------------------------------------------*/
void graphics(MESH *, DOF_REAL_VEC *, REAL (*)(EL *));
